"""
es_client Module Initialization

This module initializes the es_client package, providing utilities for building
Elasticsearch client connections and debugging tools.

Exported Objects:
    - Builder: A class to construct Elasticsearch client connections.
    - debug: A debugging utility for logging at various verbosity levels.

Example Usage:
    >>> from es_client import Builder
    >>> config = {'elasticsearch': {'client': {'hosts': ['http://localhost:9200']}}}
    >>> builder = Builder(configdict=config)
    >>> builder.client_args.hosts
    ['http://localhost:9200']
    >>> from es_client import debug
    >>> debug.lv1("Debug message")
    # Outputs debug message at level 1, if logging is configured appropriately.

Version:
    9.0.2
"""

from datetime import datetime
from .builder import Builder
from .debug import debug

FIRST_YEAR = 2022
now = datetime.now()
if now.year == FIRST_YEAR:
    COPYRIGHT_YEARS = "2025"
else:
    COPYRIGHT_YEARS = f"2025-{now.year}"

__version__ = "9.0.2"
__author__ = "Aaron Mildenstein"
__copyright__ = f"{COPYRIGHT_YEARS}, {__author__}"
__license__ = "Apache 2.0"
__status__ = "Development"
__description__ = "Elasticsearch Client builder, complete with schema validation"
__url__ = "https://github.com/untergeek/es_client"
__email__ = "aaron@mildensteins.com"
__maintainer__ = "Aaron Mildenstein"
__maintainer_email__ = f"{__email__}"
__keywords__ = ["elasticsearch", "client", "connect", "command-line"]
__classifiers__ = [
    "Development Status :: 5 - Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

__all__ = ["Builder", "debug", "__author__", "__copyright__", "__version__"]



"""Sphinx configuration for es_client documentation.

Configures Sphinx to generate documentation for the es_client package,
using autodoc, Napoleon, doctest, viewcode, and intersphinx extensions.
Imports metadata (__version__, __author__, __copyright__) from
es_client, leveraging module installation for ReadTheDocs. Sets up
GitHub integration for "Edit Source" links and supports Python 3.8-3.13.

Attributes:
    project: Project name ("es_client"). (str)
    author: Author name from es_client.__author__. (str)
    version: Major.minor version (e.g., "1.3"). (str)
    release: Full version (e.g., "1.3.0"). (str)
    html_theme: Theme for HTML output, defaults to "furo". (str)

Examples:
    >>> project
    'es_client'
    >>> author
    'Aaron Mildenstein'
    >>> version
    '1.3'
    >>> 'autodoc' in [ext.split('.')[-1] for ext in extensions]
    True
"""

# pylint: disable=C0103,E0401,W0622

# -- Imports and setup -----------------------------------------------------

from es_client import __author__, __copyright__, __version__

# -- Project information -----------------------------------------------------

project = "es_client"
github_user = "untergeek"
github_repo = "es_client"
github_branch = "master"
author = __author__
copyright = __copyright__
release = __version__
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]
napoleon_google_docstring = True
napoleon_numpy_docstring = False

templates_path = ["_templates"]
exclude_patterns = ["_build"]
source_suffix = ".rst"
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"

# Add "Edit Source" links into the template
html_context = {
    "display_github": True,
    "github_user": f"{github_user}",
    "github_repo": f"{github_repo}",
    "github_version": f"{github_branch}",
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}

# -- Autodoc configuration ---------------------------------------------------

autoclass_content = "both"
autodoc_class_signature = "separated"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
}
autodoc_member_order = "bysource"
autodoc_typehints = "description"

# -- Intersphinx configuration -----------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.13", None),
    "elasticsearch9": ("https://elasticsearch-py.readthedocs.io/en/v9.1.1", None),
    "elastic-transport": (
        "https://elastic-transport-python.readthedocs.io/en/stable",
        None,
    ),
    "voluptuous": ("http://alecthomas.github.io/voluptuous/docs/_build/html", None),
    "click": ("https://click.palletsprojects.com/en/stable", None),
}



"""Builder and associated Classes

This module provides the `Builder` class to construct Elasticsearch client connections
using configuration from dictionaries or YAML files.

Classes:
    Builder: Constructs an Elasticsearch client with validated configuration.
    SecretStore: Securely stores sensitive fields using Fernet encryption.
"""

# pylint: disable=C0415,R0902,R0913,R0917
import typing as t
import logging
import warnings
from dotmap import DotMap  # type: ignore
from cryptography.fernet import Fernet
import tiered_debug as debug
from elastic_transport import ObjectApiResponse
import elasticsearch9
from .debug import debug, begin_end
from .defaults import (
    VERSION_MIN,
    VERSION_MAX,
    CLIENT_SETTINGS,
    OTHER_SETTINGS,
    ES_DEFAULT,
)
from .exceptions import ConfigurationError, ESClientException, NotMaster
from .schemacheck import password_filter
from .utils import (
    check_config,
    ensure_list,
    file_exists,
    get_version,
    get_yaml,
    parse_apikey_token,
    prune_nones,
    verify_ssl_paths,
    verify_url_schema,
)

logger = logging.getLogger(__name__)

# Error message constants
INVALID_HOST_SCHEMA = "Invalid host schema: {host}"
MUST_PROVIDE_BOTH_AUTH = "Must populate both username and password, or neither"
MUST_PROVIDE_BOTH_API_KEY = "Must populate both id and api_key, or neither"
HOSTS_AND_CLOUD_ID_CONFLICT = 'Cannot populate both "hosts" and "cloud_id"'
MULTIPLE_HOSTS_MASTER_ONLY = (
    '"master_only" cannot be True if multiple hosts are specified. Hosts = {hosts}'
)
NOT_MASTER_NODE = (
    "master_only is True, but the client is connected to a non-master node."
)
UNSUPPORTED_VERSION = "Elasticsearch version {version} not supported"
FILE_NOT_FOUND = '"{key}: {path}" File not found!'


class SecretStore:
    """
    Securely stores secrets using Fernet encryption.

    Args:
        key (bytes, optional): Fernet key for encryption. If None, generates a key.

    Example:
        >>> store = SecretStore()
        >>> store.store_secret("api_key", ("id", "key"))
        >>> store.get_secret("api_key")
        ('id', 'key')
    """

    def __init__(self, key: t.Optional[bytes] = None):

        self._fernet = Fernet(key or Fernet.generate_key())
        self._secrets: t.Dict[str, bytes] = {}

    def store_secret(self, name: str, value: t.Any) -> None:
        """Encrypt and store a secret."""
        import json

        serialized = json.dumps(value).encode()
        self._secrets[name] = self._fernet.encrypt(serialized)

    def get_secret(self, name: str) -> t.Any:
        """Decrypt and return a secret, or None if not found."""
        import json

        if name in self._secrets:
            decrypted = self._fernet.decrypt(self._secrets[name]).decode()
            return json.loads(decrypted)
        return None

    def __repr__(self) -> str:
        """Return a safe string representation."""
        return f"<SecretStore with {len(self._secrets)} secrets>"


class Builder:
    """
    Constructs an Elasticsearch client connection from configuration.

    The `Builder` class processes configuration from a dictionary or YAML file,
    validates it, and creates an :class:`~elasticsearch9.Elasticsearch` client.
    It supports automatic connection and version checking, with options for
    master-only connections. Sensitive fields are stored securely in a
    :class:`SecretStore`.

    Parameters:
        configdict (dict, optional): Configuration dictionary with an 'elasticsearch'
            key containing 'client' and 'other_settings' subkeys. Defaults to None.
        configfile (str, optional): Path to a YAML file with the same structure as
            configdict. Defaults to None.
        autoconnect (bool, optional): Connect to client automatically. Defaults
            to False.
        version_min (tuple, optional): Minimum Elasticsearch version as (major, minor,
            patch). Defaults to :const:`~es_client.defaults.VERSION_MIN`.
        version_max (tuple, optional): Maximum Elasticsearch version as
            (major, minor, patch). Defaults to :const:`~es_client.defaults.VERSION_MAX`.

    Attributes:
        attributes (DotMap): Storage for configuration and settings.
        client (:class:`~elasticsearch9.Elasticsearch`): The Elasticsearch client
            connection.
        _secrets (:class:`SecretStore`): Secure storage for sensitive fields.

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If configuration is invalid,
            such as an invalid host schema (checked during initialization).
        :exc:`~es_client.exceptions.ESClientException`: If connection to Elasticsearch
            fails.
        :exc:`~es_client.exceptions.NotMaster`: If `master_only` is True and connected
            node is not the master.

    Examples:
        >>> config = {'elasticsearch': {'client': {'hosts': ['http://localhost:9200']}}}
        >>> builder = Builder(configdict=config)
        >>> builder.client_args.hosts
        ['http://localhost:9200']
        >>> builder.master_only = True
        >>> builder.master_only
        True
        >>> cfg = {'elasticsearch': {'client': {'hosts': ['ftp://invalid']}}})
        >>> Builder(configdict=cfg)
        Traceback (most recent call last):
            ...
        es_client.exceptions.ConfigurationError: Invalid host schema: ftp://invalid
    """

    def __init__(
        self,
        configdict: t.Union[t.Dict, None] = None,
        configfile: t.Union[str, None] = None,
        autoconnect: bool = False,
        version_min: t.Tuple = VERSION_MIN,
        version_max: t.Tuple = VERSION_MAX,
    ):
        debug.lv2('Initializing Builder object...')
        self.attributes = DotMap()
        self.config = DotMap()
        self.config.client = DotMap()
        self.config.client.hosts = []
        self.config.other_settings = DotMap()
        self._secrets = SecretStore()
        self.set_client_defaults()
        self.set_other_defaults()
        self.client = elasticsearch9.Elasticsearch(hosts="http://127.0.0.1:9200")
        self.process_config_opts(configdict, configfile)
        # Validate host schemas immediately
        if self.config.client.get("hosts"):
            verified_hosts = []
            for host in ensure_list(self.config.client["hosts"]):
                try:
                    debug.lv4(f'TRY: validate host {host}')
                    verified_hosts.append(verify_url_schema(host))
                except ConfigurationError as exc:
                    logger.critical(INVALID_HOST_SCHEMA.format(host=host))
                    debug.lv3('Exiting method, raising exception')
                    debug.lv5(f'Exception = "{exc}"')
                    raise ConfigurationError(
                        INVALID_HOST_SCHEMA.format(host=host)
                    ) from exc
            self.config.client["hosts"] = verified_hosts
        self.version_max = version_max
        self.version_min = version_min
        self.update_config()
        self.validate()
        if autoconnect:
            self.connect()
            self.test_connection()
        debug.lv3('Builder object initialized')

    def __repr__(self) -> str:
        """
        Return a string representation of the Builder instance.

        Returns:
            str: A string describing the Builder's configuration and client status.

        Example:
            >>> config = {
            ...     'elasticsearch': {
            ...         'client': {
            ...             'hosts': ['http://localhost:9200'],
            ...             'cloud_id': 'my_cloud_id'
            ...         }
            ...     }
            ... }
            >>> builder = Builder(configdict=config)
            >>> repr(builder)  # doctest: +ELLIPSIS
            "Builder(hosts=['http://localhost:9200'], master_only=False,
            version_min=(9, 0, 0), cloud_id='my_cloud_id', ...)"
            >>> config = {
                'elasticsearch': {
                    'client': {'hosts': ['http://localhost:9200']}
                }
            }
            >>> builder = Builder(configdict=config)
            >>> repr(builder)  # doctest: +ELLIPSIS
            "Builder(hosts=['http://localhost:9200'], master_only=False,
            version_min=(9, 0, 0), ...)"
        """
        hosts = self.client_args.hosts or ['None']
        base = (
            f"Builder(hosts={hosts}, master_only={self.master_only}, "
            f"version_min={self.version_min}"
        )
        if self.client_args.cloud_id:
            base += f", cloud_id='{self.client_args.cloud_id}'"
        base += f", client={self.client})"
        return base

    @property
    def master_only(self) -> bool:
        """
        Get or set whether to connect only to the elected master node.

        Returns:
            bool: True if only the master node is allowed, False otherwise.

        Example:
            >>> builder = Builder()
            >>> builder.master_only = True
            >>> builder.master_only
            True
        """
        return self.attributes.master_only

    @master_only.setter
    def master_only(self, value: bool) -> None:
        self.attributes.master_only = value

    @property
    def is_master(self) -> bool:
        """
        Get or set whether the connected node is the elected master.

        Returns:
            bool: True if the connected node is the master, False otherwise.

        Example:
            >>> builder = Builder()
            >>> builder.is_master = False
            >>> builder.is_master
            False
        """
        return self.attributes.is_master

    @is_master.setter
    def is_master(self, value: bool) -> None:
        self.attributes.is_master = value

    @property
    def config(self) -> DotMap:
        """
        Get or set the configuration settings from configfile or configdict.

        Returns:
            DotMap: Configuration settings.

        Example:
            >>> config = {
            ...     'elasticsearch': {
            ...         'client': {'hosts': ['http://localhost:9200']}
            ...     }
            ... }
            >>> builder = Builder(configdict=config)
            >>> builder.config.client.hosts
            ['http://localhost:9200']
        """
        return self.attributes.config

    @config.setter
    def config(self, value: t.Dict) -> None:
        self.attributes.config = DotMap(value)

    @property
    def client_args(self) -> DotMap:
        """
        Get or set the client settings.

        Returns:
            DotMap: Client configuration settings.

        Example:
            >>> builder = Builder()
            >>> builder.client_args.hosts = ['http://localhost:9200']
            >>> builder.client_args.hosts
            ['http://localhost:9200']
        """
        return self.attributes.client_args

    @client_args.setter
    def client_args(self, value: t.Dict) -> None:
        self.attributes.client_args = DotMap(value)

    @property
    def other_args(self) -> DotMap:
        """
        Get or set the other settings.

        Returns:
            DotMap: Other configuration settings.

        Example:
            >>> builder = Builder()
            >>> builder.other_args.master_only = True
            >>> builder.other_args.master_only
            True
        """
        return self.attributes.other_args

    @other_args.setter
    def other_args(self, value: t.Dict) -> None:
        self.attributes.other_args = DotMap(value)

    @property
    def skip_version_test(self) -> bool:
        """
        Get or set whether to skip version compatibility tests.

        Returns:
            bool: True if version tests are skipped, False otherwise.

        Example:
            >>> builder = Builder()
            >>> builder.skip_version_test = True
            >>> builder.skip_version_test
            True
        """
        return self.attributes.skip_version_test

    @skip_version_test.setter
    def skip_version_test(self, value: bool) -> None:
        self.attributes.skip_version_test = value

    @property
    def version_min(self) -> t.Tuple:
        """
        Get or set the minimum acceptable Elasticsearch version.

        Returns:
            tuple: Minimum version as (major, minor, patch).

        Example:
            >>> builder = Builder()
            >>> builder.version_min
            (9, 0, 0)
        """
        return self.attributes.version_min

    @version_min.setter
    def version_min(self, value: t.Tuple) -> None:
        self.attributes.version_min = value

    @property
    def version_max(self) -> t.Tuple:
        """
        Get or set the maximum acceptable Elasticsearch version.

        Returns:
            tuple: Maximum version as (major, minor, patch).

        Example:
            >>> builder = Builder()
            >>> builder.version_max
            (9, 99, 99)
        """
        return self.attributes.version_max

    @version_max.setter
    def version_max(self, value: t.Tuple) -> None:
        self.attributes.version_max = value

    @begin_end()
    def set_client_defaults(self) -> None:
        """
        Set default values for client_args.

        Initializes client_args with None for all keys in
        :const:`~es_client.defaults.CLIENT_SETTINGS`.
        """
        self.client_args = DotMap()
        for key in CLIENT_SETTINGS:
            self.client_args[key] = None

    @begin_end()
    def set_other_defaults(self) -> None:
        """
        Set default values for other_args.

        Initializes other_args with None for all keys in
        :const:`~es_client.defaults.OTHER_SETTINGS`.
        """
        self.other_args = DotMap()
        for key in OTHER_SETTINGS:
            self.other_args[key] = None

    @begin_end()
    def process_config_opts(
        self, configdict: t.Union[t.Dict, None], configfile: t.Union[str, None]
    ) -> None:
        """
        Process configuration from configdict or configfile.

        Args:
            configdict (dict, optional): Configuration dictionary with an
                'elasticsearch' key containing 'client' and 'other_settings' subkeys.
            configfile (str, optional): Path to a YAML file with the same structure as
                configdict.

        Prioritizes configdict over configfile. If neither is provided, uses
        :const:`~es_client.defaults.ES_DEFAULT`, which sets hosts to
        'http://127.0.0.1:9200'.

        Example:
            >>> builder = Builder()
            >>> builder.config.client.hosts
            ['http://127.0.0.1:9200']
        """
        if configfile:
            debug.lv2(f'Using values from file: {configfile}')
            self.config = check_config(get_yaml(configfile))
        elif configdict:
            debug.lv2(f'Using values from dict: {password_filter(configdict)}')
            self.config = check_config(configdict)
        else:
            debug.lv2("No configuration provided. Using ES_DEFAULT.")
            self.config = check_config(ES_DEFAULT["elasticsearch"])

    @begin_end()
    def update_config(self) -> None:
        """
        Update object with configuration values.

        Applies settings from config to client_args and other_args, moves sensitive
        fields to a secure store, and sets master_only and skip_version_test.
        """
        self.client_args.update(self.config.client)
        self.other_args.update(self.config.other_settings)
        # Move sensitive fields to SecretStore
        sensitive_fields = ['basic_auth', 'api_key', 'bearer_auth']
        for field in sensitive_fields:
            # We are checking in client_args in case someone has manually passed
            # something. We build basic_auth from user/pass, and api_key comes
            # from self.other_args. In other words, this is a failsafe.
            if field in self.client_args and self.client_args[field] is not None:
                self._secrets.store_secret(field, self.client_args[field])
                self.client_args[field] = None
                self.config.client[field] = None
        if 'password' in self.other_args and self.other_args.password is not None:
            self._secrets.store_secret('password', self.other_args.password)
            self.other_args.password = None
            self.config.other_settings.password = None
        self.master_only = self.other_args.master_only
        self.is_master = False
        if "skip_version_test" in self.other_args:
            self.skip_version_test = self.other_args.skip_version_test
        else:
            self.skip_version_test = False

    @begin_end()
    def validate(self) -> None:
        """
        Validate configuration settings.

        Checks basic auth, API key, cloud ID, and SSL settings. Host schemas are
        validated in :meth:`__init__` using :func:`~es_client.utils.verify_url_schema`.
        Issues warnings for experimental options like ssl_version.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If validation fails, such
                as missing authentication credentials or invalid cloud ID.

        Example:
            >>> config = {'elasticsearch': {'client': {'ssl_version': 'TLSv1'}}}
            >>> builder = Builder(configdict=config)  # doctest: +ELLIPSIS
            ... # Warning: ssl_version is experimental; use ssl_context instead
        """
        if self.client_args.ssl_version:
            warnings.warn(
                "ssl_version is experimental; use ssl_context instead",
                DeprecationWarning,
                stacklevel=2,
            )
        if self.client_args.ssl_assert_fingerprint:
            warnings.warn(
                "ssl_assert_fingerprint is experimental on CPython 3.10+; "
                "use ssl_context instead",
                DeprecationWarning,
                stacklevel=2,
            )
        self._check_basic_auth()
        self._check_api_key()
        self._check_cloud_id()
        self._check_ssl()

    @begin_end()
    def connect(self) -> None:
        """
        Establish connection to Elasticsearch.

        Performs post-connection checks for version and master status using
        :meth:`~Builder._check_version` and :meth:`~Builder._find_master`.

        Raises:
            :exc:`~es_client.exceptions.NotMaster`: If master_only is True and node is
                not master.
            :exc:`~es_client.exceptions.ESClientException`: If version is incompatible.
        """
        self._get_client()
        self._check_version()
        if self.master_only:
            self._check_multiple_hosts()
            self._find_master()
            self._check_if_master()

    @begin_end()
    def _check_basic_auth(self) -> None:
        """
        Validate and set basic authentication credentials.

        Creates basic_auth tuple from username and password if both are provided.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If only one of username or
                password is provided.
        """
        if "username" in self.other_args or "password" in self.other_args:
            usr = self.other_args.username if "username" in self.other_args else None
            pwd = self._secrets.get_secret('password')
            if usr is None and pwd is None:
                pass
            elif usr is None or pwd is None:
                debug.lv3('Exiting method, raising exception')
                debug.lv5(f'Exception = "{MUST_PROVIDE_BOTH_AUTH}"')
                raise ConfigurationError(MUST_PROVIDE_BOTH_AUTH)
            else:
                self._secrets.store_secret('basic_auth', (usr, pwd))

    @begin_end()
    def _check_api_key(self) -> None:
        """
        Validate and set API key credentials.

        Processes API key from a token or id/api_key pair in other_args.api_key.
        Token takes precedence over id and api_key.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If id or api_key is
                missing when required.

        Example:
            >>> builder = Builder()
            >>> builder.other_args.api_key = {'id': 'test_id', 'api_key': 'test_key'}
            >>> builder._check_api_key()
            >>> builder._secrets.get_secret('api_key')
            ('test_id', 'test_key')
        """
        if "api_key" not in self.other_args:
            return
        api_key_config = DotMap(self.other_args.api_key)
        if api_key_config.get("token"):
            api_id, api_key = parse_apikey_token(api_key_config.token)
            self._secrets.store_secret('api_key', (api_id, api_key))
            # Clean up sensitive fields from places they could still exist
            self.other_args.api_key.token = None
            self.config.other_settings.api_key.token = None
            return
        api_id = api_key_config.get("id")
        api_key = api_key_config.get("api_key")
        if api_id is None and api_key is None:
            self._secrets.store_secret('api_key', None)
        elif api_id is None or api_key is None:
            raise ConfigurationError(MUST_PROVIDE_BOTH_API_KEY)
        else:
            self._secrets.store_secret('api_key', (api_id, api_key))
            # Clean up sensitive fields from places they could still exist
            self.other_args.api_key.id = None
            self.config.other_settings.api_key.id = None
            self.other_args.api_key.api_key = None
            self.config.other_settings.api_key.api_key = None

    @begin_end()
    def _check_cloud_id(self) -> None:
        """
        Validate cloud_id configuration.

        Removes hosts if cloud_id is provided, as they are mutually exclusive.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If both hosts and cloud_id
                are specified.
        """
        if "cloud_id" in self.client_args and self.client_args.cloud_id is not None:
            if (
                self.client_args.hosts == ["http://127.0.0.1:9200"]
                and len(self.client_args.hosts) == 1
            ):
                self.client_args.hosts = None
            if self.client_args.hosts is not None:
                debug.lv3('Exiting method, raising exception')
                logger.error(HOSTS_AND_CLOUD_ID_CONFLICT)
                raise ConfigurationError(HOSTS_AND_CLOUD_ID_CONFLICT)

    @begin_end()
    def _check_ssl(self) -> None:
        """
        Validate SSL configuration.

        Uses certifi for HTTPS if ca_certs is not specified. Checks file existence with
        :func:`~es_client.utils.file_exists`.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If SSL certificate or key
                files are not found.
        """
        verify_ssl_paths(self.client_args)
        if "cloud_id" in self.client_args and self.client_args.cloud_id is not None:
            scheme = "https"
        elif self.client_args.hosts is None:
            scheme = None
        else:
            scheme = self.client_args.hosts[0].split(":")[0].lower()
        if scheme == "https":
            if "ca_certs" not in self.client_args or not self.client_args.ca_certs:
                import certifi

                self.client_args.ca_certs = certifi.where()
            else:
                keylist = ["ca_certs", "client_cert", "client_key"]
                for key in keylist:
                    if key in self.client_args and self.client_args[key]:
                        if not file_exists(self.client_args[key]):
                            msg = FILE_NOT_FOUND.format(
                                key=key, path=self.client_args[key]
                            )
                            logger.critical(msg)
                            debug.lv3('Exiting method, raising exception')
                            debug.lv5(f'Exception = "{msg}"')
                            raise ConfigurationError(msg)

    @begin_end()
    def _find_master(self) -> None:
        """
        Check if the connected node is the elected master.

        Sets is_master based on node ID comparison using
        :meth:`~elasticsearch9.Elasticsearch.nodes.info`.
        """
        my_node_id = list(self.client.nodes.info(node_id="_local")["nodes"])[0]
        master_node_id = self.client.cluster.state(metric="master_node")["master_node"]
        self.is_master = my_node_id == master_node_id

    @begin_end()
    def _check_multiple_hosts(self) -> None:
        """
        Validate host count for master_only setting.

        Raises:
            :exc:`~es_client.exceptions.ConfigurationError`: If master_only is True and
                multiple hosts are specified.
        """
        if "hosts" in self.client_args and isinstance(self.client_args.hosts, list):
            if len(self.client_args.hosts) > 1:
                debug.lv3('Exiting method, raising exception')
                hosts = self.client_args.hosts
                msg = MULTIPLE_HOSTS_MASTER_ONLY.format(hosts=hosts)
                logger.error(msg)
                raise ConfigurationError(msg)

    @begin_end()
    def _check_if_master(self) -> None:
        """
        Verify if connected node is the master when master_only is True.

        Raises:
            :exc:`~es_client.exceptions.NotMaster`: If connected node is not the master.
        """
        if not self.is_master:
            debug.lv3('Exiting method, raising exception')
            logger.error(NOT_MASTER_NODE)
            raise NotMaster(NOT_MASTER_NODE)

    @begin_end()
    def _check_version(self) -> None:
        """
        Verify Elasticsearch version compatibility.

        Compares cluster version against version_min and version_max using
        :func:`~es_client.utils.get_version`.

        Raises:
            :exc:`~es_client.exceptions.ESClientException`: If version is outside
                acceptable range.
        """
        v = get_version(self.client)
        if self.skip_version_test:
            logger.warning("Skipping Elasticsearch version checks")
        else:
            debug.lv2(f'Version detected: {".".join(map(str, v))}')
            if v >= self.version_max or v < self.version_min:
                msg = UNSUPPORTED_VERSION.format(version='.'.join(map(str, v)))
                debug.lv3('Exiting method, raising exception')
                logger.error(msg)
                raise ESClientException(msg)

    @begin_end()
    def _get_client(self) -> None:
        """
        Instantiate the :class:`~elasticsearch9.Elasticsearch` client.

        Creates client with pruned configuration arguments using
        :func:`~es_client.utils.prune_nones`, including sensitive fields from
        the secure store.
        """
        client_args = prune_nones(self.client_args.toDict())
        # Add sensitive fields from SecretStore
        for field in ['basic_auth', 'api_key', 'bearer_auth']:
            secret = self._secrets.get_secret(field)
            if secret is not None and field == 'basic_auth':
                client_args[field] = tuple(secret)
            elif secret is not None:
                client_args[field] = secret
        self.client = elasticsearch9.Elasticsearch(**client_args)

    @begin_end()
    def test_connection(self) -> ObjectApiResponse[t.Any]:
        """
        Test the Elasticsearch connection.

        Executes :meth:`~elasticsearch9.Elasticsearch.info` to verify connectivity.

        Returns:
            :class:`~elastic_transport.ObjectApiResponse`: Response from Elasticsearch
                info API.
        """
        retval = self.client.info()
        debug.lv5(f'Return value = "{retval}"')
        return retval



"""Utility functions for es_client

This module provides helper functions for configuration validation, file handling, and
CLI setup in es_client. It supports :class:`~es_client.builder.Builder`,
:mod:`~es_client.logging`, and :mod:`~es_client.config` with tasks like URL validation,
YAML parsing, and sensitive data redaction.

Functions:
    check_config: Validate an Elasticsearch configuration dictionary.
    ensure_list: Convert a scalar or list to a list.
    file_exists: Check if a file exists.
    get_version: Retrieve the Elasticsearch version as a tuple.
    get_yaml: Load a YAML file with environment variable parsing.
    option_wrapper: Wrap click.option for configuration storage.
    parse_apikey_token: Decode a base64-encoded API key token.
    passthrough: Generic wrapper for click decorators.
    prune_nones: Remove keys with None values from a dictionary.
    read_file: Read a file's contents.
    verify_ssl_paths: Verify readability of SSL certificate/key paths.
    verify_url_schema: Validate and normalize URL schemas for hosts.
"""

import typing as t
import logging
import os
import re
import base64
import binascii
from pathlib import Path
from urllib.parse import urlparse
import yaml  # type: ignore
from yaml.scanner import ScannerError  # type: ignore
from yaml.parser import ParserError  # type: ignore
import click
from elasticsearch9 import Elasticsearch
from .debug import debug, begin_end
from .defaults import ES_DEFAULT, config_schema
from .exceptions import ConfigurationError
from .schemacheck import SchemaCheck, password_filter

logger = logging.getLogger(__name__)


@begin_end()
def check_config(config: dict, quiet: bool = False) -> dict:
    """
    Validate an Elasticsearch configuration dictionary.

    Args:
        config (dict): Configuration dictionary to validate.
        quiet (bool, optional): Suppress warning logs if True. Defaults to False.

    Returns:
        dict: Validated configuration for :class:`~es_client.builder.Builder`.

    Ensures 'elasticsearch', 'client', and 'other_settings' keys are present in
    `config`, using :data:`~es_client.defaults.ES_DEFAULT` if missing. Validates
    against :func:`~es_client.defaults.config_schema` via
    :class:`~es_client.schemacheck.SchemaCheck`.

    Raises:
        :exc:`~es_client.exceptions.FailedValidation`: If validation fails.

    Example:
        >>> config = {'elasticsearch': {'client': {'hosts': ['http://localhost:9200']}}}
        >>> result = check_config(config, quiet=True)
        >>> result['client']['hosts']
        ['http://localhost:9200']
        >>> check_config([], quiet=True)
        {'client': {}, 'other_settings': {}}
    """
    if not isinstance(config, dict):
        logger.warning(
            "Elasticsearch client configuration must be provided as a dictionary."
        )
        logger.warning('You supplied: "%s" which is "%s".', config, type(config))
        logger.warning("Using default values.")
        es_settings = ES_DEFAULT
    elif "elasticsearch" not in config:
        if not quiet:
            logger.warning(
                'No "elasticsearch" setting in configuration. Using defaults.'
            )
        es_settings = ES_DEFAULT
    else:
        es_settings = config
    for key in ["client", "other_settings"]:
        if key not in es_settings["elasticsearch"]:
            es_settings["elasticsearch"][key] = {}
        else:
            es_settings["elasticsearch"][key] = prune_nones(
                es_settings["elasticsearch"][key]
            )
    _ = SchemaCheck(
        es_settings["elasticsearch"],
        config_schema(),
        "Elasticsearch Configuration",
        "elasticsearch",
    ).result()
    retval = dict(_)
    debug.lv5(f'Return value = "{password_filter(retval)}"')
    return retval


@begin_end()
def ensure_list(data: t.Any) -> list:
    """
    Convert a scalar or list to a list.

    Args:
        data: A scalar or list to convert.

    Returns:
        list: A list containing `data` if scalar, or `data` itself if already a list.

    Example:
        >>> ensure_list('item')
        ['item']
        >>> ensure_list(['item1', 'item2'])
        ['item1', 'item2']
    """
    if not isinstance(data, list):
        data = [data]
    debug.lv5(f'Return value = "{data}"')
    return data


@begin_end()
def file_exists(file: str) -> bool:
    """
    Check if a file exists.

    Args:
        file (str): Path to the file.

    Returns:
        bool: True if `file` exists, False otherwise.

    Example:
        >>> from pathlib import Path
        >>> Path('test.txt').write_text('test')
        4
        >>> file_exists('test.txt')
        True
        >>> file_exists('nonexistent.txt')
        False
        >>> Path('test.txt').unlink()
    """
    retval = Path(file).is_file()
    debug.lv5(f'Return value = "{retval}"')
    return retval


@begin_end()
def get_version(client: Elasticsearch) -> t.Tuple:
    """
    Retrieve the Elasticsearch version as a tuple.

    Args:
        client (:class:`elasticsearch9.Elasticsearch`): Elasticsearch client instance.

    Returns:
        tuple: Version as (major, minor, patch).

    Extracts the version from `client.info()['version']['number']`, stripping
    non-semantic tags (e.g., '-dev', '-beta') and limiting to three parts.

    Example:
        >>> from unittest.mock import Mock
        >>> client = Mock()
        >>> client.info.return_value = {'version': {'number': '9.0.0'}}
        >>> get_version(client)
        (9, 0, 0)
    """
    version = client.info()["version"]["number"]
    version = version.split("-")[0]
    if len(version.split(".")) > 3:
        version = version.split(".")[:-1]
    else:
        version = version.split(".")
    retval = tuple(map(int, version))
    debug.lv5(f'Return value = "{retval}"')
    return retval


@begin_end()
def get_yaml(path: str) -> t.Dict:
    """
    Load a YAML file with environment variable parsing.

    Args:
        path (str): Path to the YAML file.

    Returns:
        dict: Parsed YAML contents.

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If the YAML file is invalid
            or unreadable.

    Supports environment variable substitution in YAML using `${VAR:DEFAULT}` syntax.

    Example:
        >>> from pathlib import Path
        >>> Path('test.yaml').write_text('host: ${HOST:localhost}')
        23
        >>> os.environ['HOST'] = '127.0.0.1'
        >>> config = get_yaml('test.yaml')
        >>> config['host']
        '127.0.0.1'
        >>> Path('test.yaml').unlink()
    """
    single = re.compile(r"^\$\{(.*)\}$")
    yaml.add_implicit_resolver("!single", single)

    def single_constructor(loader, node):
        value = loader.construct_scalar(node)
        proto = single.match(value).group(1)  # type: ignore[union-attr]
        default = None
        if len(proto.split(":")) > 1:
            envvar, default = proto.split(":")
        else:
            envvar = proto
        return os.environ[envvar] if envvar in os.environ else default

    yaml.add_constructor("!single", single_constructor)
    try:
        debug.lv4('TRY: yaml.load()')
        retval = yaml.load(read_file(path), Loader=yaml.FullLoader)
    except (ScannerError, ParserError) as exc:
        raise ConfigurationError(f"Unable to parse YAML file. Error: {exc}") from exc
    debug.lv5('Return value = "<REDACTED YAML>"')
    return retval


def option_wrapper() -> t.Callable:
    """
    Wrap click.option for configuration storage.

    Returns:
        callable: Wrapper function for :func:`click.option`.

    Example:
        >>> wrapper = option_wrapper()
        >>> isinstance(wrapper(('option',), {}), click.decorators._Option)
        True
    """
    return passthrough(click.option)


@begin_end()
def parse_apikey_token(token: str) -> t.Tuple:
    """
    Decode a base64-encoded API key token.

    Args:
        token (str): Base64-encoded token in the format 'id:api_key'.

    Returns:
        tuple: (id, api_key) extracted from the token.

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If the token is invalid.

    Example:
        >>> parse_apikey_token('Zm9vOmJhcg==')  # base64 for 'foo:bar'
        ('foo', 'bar')
        >>> parse_apikey_token('invalid')
        Traceback (most recent call last):
            ...
        es_client.exceptions.ConfigurationError: Unable to parse base64 API Key
        Token: Incorrect padding
    """
    try:
        debug.lv4('TRY: base64.b64decode()')
        decoded = base64.b64decode(token).decode("utf-8")
        split = decoded.split(":")
    except (binascii.Error, IndexError, UnicodeDecodeError) as exc:
        debug.lv3('Exiting function, raising exception')
        debug.lv5(f'Value = "{exc}"')
        logger.error(
            "Unable to parse base64 API Key Token. Ensure you are using the correct "
            "format: <id>:<api_key>"
        )
        raise ConfigurationError(
            f"Unable to parse base64 API Key Token: {exc}"
        ) from exc
    retval = (split[0], split[1])
    debug.lv5('Return value = "<REDACTED API Key>"')
    return retval


def passthrough(func) -> t.Callable:
    """
    Generic wrapper for click decorators.

    Args:
        func (callable): Function to wrap (e.g., :func:`click.option`).

    Returns:
        callable: Wrapped function passing arguments through.

    Example:
        >>> wrapped = passthrough(lambda x: x * 2)
        >>> wrapped(5)
        10
    """
    return lambda a, k: func(*a, **k)


@begin_end()
def prune_nones(mydict: t.Dict) -> t.Dict:
    """
    Remove keys with None values from a dictionary.

    Args:
        mydict (dict): Dictionary to process.

    Returns:
        dict: Dictionary with None-valued keys removed.

    Example:
        >>> prune_nones({'a': 1, 'b': None, 'c': 'None'})
        {'a': 1}
    """
    retval = dict([(k, v) for k, v in mydict.items() if v is not None and v != "None"])
    debug.lv5(f'Return value = "{password_filter(retval)}"')
    return retval


@begin_end()
def read_file(myfile: str) -> str:
    """
    Read a file's contents.

    Args:
        myfile (str): Path to the file.

    Returns:
        str: File contents.

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If the file cannot be read.

    Example:
        >>> from pathlib import Path
        >>> Path('test.txt').write_text('test')
        4
        >>> read_file('test.txt')
        'test'
        >>> Path('test.txt').unlink()
    """
    try:
        debug.lv4('TRY: open() and read() file...')
        with open(myfile, "r", encoding="utf-8") as f:
            data = f.read()
        debug.lv3('Exiting function, returning value')
        debug.lv5('Value = "<REDACTED FILE>"')
        return data
    except IOError as exc:
        msg = f"Unable to read file {myfile}. Exception: {exc}"
        debug.lv3('Exiting function, raising exception')
        logger.error(msg)
        raise ConfigurationError(msg) from exc


@begin_end()
def verify_ssl_paths(args: t.Dict) -> None:
    """
    Verify readability of SSL certificate/key paths.

    Args:
        args (dict): Configuration with 'ca_certs', 'client_cert', or 'client_key' keys.

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If any file is unreadable.

    Checks if paths in `args` for 'ca_certs', 'client_cert', or 'client_key' are
    readable using :func:`read_file`.

    Example:
        >>> from pathlib import Path
        >>> Path('cert.pem').write_text('cert')
        4
        >>> verify_ssl_paths({'ca_certs': 'cert.pem'})
        >>> verify_ssl_paths({'ca_certs': 'nonexistent.pem'})
        Traceback (most recent call last):
            ...
        es_client.exceptions.ConfigurationError: Unable to read file nonexistent.pem.
        Exception: [Errno 2] No such file or directory: 'nonexistent.pem'
        >>> Path('cert.pem').unlink()
    """
    if "ca_certs" in args and args["ca_certs"] is not None:
        read_file(args["ca_certs"])
    if "client_cert" in args and args["client_cert"] is not None:
        read_file(args["client_cert"])
    if "client_key" in args and args["client_key"] is not None:
        read_file(args["client_key"])


@begin_end()
def verify_url_schema(url: str) -> str:
    """
    Validate and normalize a URL schema for Elasticsearch hosts.

    Args:
        url (str): URL to validate (e.g., 'http://localhost:9200',
            'https://example.com/path').

    Returns:
        str: Normalized URL with schema and port (e.g., 'http://localhost:80').

    Raises:
        :exc:`~es_client.exceptions.ConfigurationError`: If the URL schema is invalid.

    Ensures the URL uses 'http' or 'https' and includes a port (defaults to 80 for
    http, 443 for https if omitted). Accepts optional paths after the domain.

    Example:
        >>> verify_url_schema('https://localhost')
        'https://localhost:443'
        >>> verify_url_schema('http://localhost:9200/path')
        'http://localhost:9200'
        >>> verify_url_schema('ftp://localhost')
        Traceback (most recent call last):
            ...
        es_client.exceptions.ConfigurationError: URL Schema invalid for ftp://localhost
    """
    try:
        parsed = urlparse(url)
    except Exception as exc:
        errmsg = f"URL Schema invalid for {url}"
        debug.lv3('Exiting function, raising exception')
        debug.lv5(f'Value = "{errmsg}"')
        logger.error(f'Invalid URL format: "{url}"')
        raise ConfigurationError(errmsg) from exc

    if parsed.scheme not in ('http', 'https'):
        errmsg = f"URL Schema invalid for {url}"
        debug.lv3('Exiting function, raising exception')
        debug.lv5(f'Value = "{errmsg}"')
        logger.error(f'Invalid URL schema: "{url}". Only http and https are supported.')
        raise ConfigurationError(errmsg)

    if not parsed.hostname:
        errmsg = f"URL Schema invalid for {url}"
        debug.lv3('Exiting function, raising exception')
        debug.lv5(f'Value = "{errmsg}"')
        logger.error(f'Invalid URL: "{url}". Missing hostname.')
        raise ConfigurationError(errmsg)

    # Use provided port or default based on scheme
    if parsed.port is not None:
        port = str(parsed.port)
    elif parsed.scheme == "https":
        port = "443"
    else:  # http
        port = "80"

    retval = f"{parsed.scheme}://{parsed.hostname}:{port}"
    debug.lv5(f'Return value = "{retval}"')
    return retval



"""Define default values"""

# pylint: disable=line-too-long
import typing as t
from copy import deepcopy
from click import Choice, Path
from voluptuous import All, Any, Boolean, Coerce, Optional, Range, Schema

VERSION_MIN: t.Tuple = (9, 0, 0)
"""Minimum compatible Elasticsearch version"""

VERSION_MAX: t.Tuple = (9, 99, 99)
"""Maximum compatible Elasticsearch version"""

KEYS_TO_REDACT: t.List[str] = [
    "password",
    "basic_auth",
    "bearer_auth",
    "api_key",
    "id",
    "opaque_id",
    "token",
]
"""
When doing configuration Schema validation, redact the value from any listed dictionary
key. This only happens if logging is at DEBUG level.
"""

CLIENT_SETTINGS: t.List[str] = [
    "hosts",
    "cloud_id",
    "api_key",
    "basic_auth",
    "bearer_auth",
    "opaque_id",
    "headers",
    "connections_per_node",
    "http_compress",
    "verify_certs",
    "ca_certs",
    "client_cert",
    "client_key",
    "ssl_assert_hostname",
    "ssl_assert_fingerprint",
    "ssl_version",
    "ssl_context",
    "ssl_show_warn",
    "transport_class",
    "request_timeout",
    "node_class",
    "node_pool_class",
    "randomize_nodes_in_pool",
    "node_selector_class",
    "dead_node_backoff_factor",
    "max_dead_node_backoff",
    "serializer",
    "serializers",
    "default_mimetype",
    "max_retries",
    "retry_on_status",
    "retry_on_timeout",
    "sniff_on_start",
    "sniff_before_requests",
    "sniff_on_node_failures",
    "sniff_timeout",
    "min_delay_between_sniffing",
    "sniffed_node_callback",
    "meta_header",
    "host_info_callback",
    "_transport",
]
"""
Valid argument/option names for :py:class:`~.elasticsearch9.Elasticsearch`. Too large
to show
"""

OTHER_SETTINGS: t.List[str] = [
    "master_only",
    "skip_version_test",
    "username",
    "password",
    "api_key",
]
"""Valid option names for :py:class:`~.es_client.builder.Builder`'s other settings"""

CLICK_SETTINGS: t.Dict[str, t.Dict] = {
    "config": {"help": "Path to configuration file.", "type": Path(exists=True)},
    "hosts": {"help": "Elasticsearch URL to connect to.", "multiple": True},
    "cloud_id": {"help": "Elastic Cloud instance id"},
    "api_token": {"help": "The base64 encoded API Key token", "type": str},
    "id": {"help": 'API Key "id" value', "type": str},
    "api_key": {"help": 'API Key "api_key" value', "type": str},
    "username": {"help": "Elasticsearch username", "type": str},
    "password": {"help": "Elasticsearch password", "type": str},
    "bearer_auth": {"help": "Bearer authentication token", "type": str, "hidden": True},
    "opaque_id": {"help": "X-Opaque-Id HTTP header value", "type": str, "hidden": True},
    "request_timeout": {"help": "Request timeout in seconds", "type": float},
    "http_compress": {
        "help": "Enable HTTP compression",
        "default": None,
        "hidden": True,
    },
    "verify_certs": {"help": "Verify SSL/TLS certificate(s)", "default": None},
    "ca_certs": {"help": "Path to CA certificate file or directory", "type": str},
    "client_cert": {"help": "Path to client certificate file", "type": str},
    "client_key": {"help": "Path to client key file", "type": str},
    "ssl_assert_hostname": {
        "help": "Hostname or IP address to verify on the node's certificate.",
        "type": str,
        "hidden": True,
    },
    "ssl_assert_fingerprint": {
        "help": (
            "SHA-256 fingerprint of the node's certificate. If this value is given "
            "then root-of-trust verification isn't done and only the node's "
            "certificate fingerprint is verified."
        ),
        "type": str,
        "hidden": True,
    },
    "ssl_version": {
        "help": "Minimum acceptable TLS/SSL version",
        "type": str,
        "hidden": True,
    },
    "master-only": {
        "help": "Only run if the single host provided is the elected master",
        "default": None,
        "hidden": True,
    },
    "skip_version_test": {
        "help": "Elasticsearch version compatibility check",
        "default": None,
        "hidden": True,
    },
}
"""Default settings used for building :py:class:`click.Option`. Too large to show."""

ES_DEFAULT: t.Dict = {"elasticsearch": {"client": {"hosts": ["http://127.0.0.1:9200"]}}}
"""Default settings for :py:class:`~.es_client.builder.Builder`"""

ENV_VAR_PREFIX: str = "ESCLIENT"
"""Environment variable prefix"""

LOGLEVEL: None = None
"""Default loglevel"""

LOGFILE: None = None
"""Default value for logfile"""

LOGFORMAT: t.Literal['default', 'ecs', 'json'] = 'default'
"""Default value for logformat"""

BLACKLIST: None = None
"""Default value for logging blacklist"""

LOGDEFAULTS: t.Dict = {
    "loglevel": LOGLEVEL,
    "logfile": LOGFILE,
    "logformat": LOGFORMAT,
    "blacklist": BLACKLIST,
}
"""All logging defaults in a single combined dictionary"""

LOGGING_SETTINGS: t.Dict[str, t.Dict] = {
    "loglevel": {
        "help": "Log level",
        "type": Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
        "default": None,
    },
    "logfile": {"help": "Log file", "type": str},
    "logformat": {
        "help": "Log output format",
        "type": Choice(["default", "json", "ecs"]),
        "default": None,
    },
    "blacklist": {
        "help": "Named entities will not be logged",
        "multiple": True,
        "default": None,
        "hidden": True,
    },
}
"""
Default logging settings used for building :py:class:`click.Option`. Too large to show.
"""

SHOW_OPTION: t.Dict[str, bool] = {"hidden": False}
"""Override value to "unhide" a :py:class:`click.Option`"""

SHOW_ENVVAR: t.Dict[str, bool] = {"show_envvar": True}
"""Override value to make Click's help output show the associated environment variable
"""

OVERRIDE: t.Dict = {**SHOW_OPTION, **SHOW_ENVVAR}
"""Override value to combine these into a single constant"""

ONOFF: t.Dict[str, str] = {"on": "", "off": "no-"}
"""Default values for enable/disable click options"""

OPTION_DEFAULTS: t.Dict[str, t.Dict] = {
    "config": {},
    "hosts": {},
    "cloud_id": {},
    "api_token": {},
    "id": {},
    "api_key": {},
    "username": {},
    "password": {},
    "bearer_auth": {},
    "opaque_id": {},
    "request_timeout": {},
    "http_compress": {"onoff": ONOFF},
    "verify_certs": {"onoff": ONOFF},
    "ca_certs": {},
    "client_cert": {},
    "client_key": {},
    "ssl_assert_hostname": {},
    "ssl_assert_fingerprint": {},
    "ssl_version": {},
    "master-only": {"onoff": ONOFF},
    "skip_version_test": {"onoff": ONOFF},
    "loglevel": {"settings": LOGGING_SETTINGS["loglevel"]},
    "logfile": {"settings": LOGGING_SETTINGS["logfile"]},
    "logformat": {"settings": LOGGING_SETTINGS["logformat"]},
    "blacklist": {"settings": LOGGING_SETTINGS["blacklist"]},
}
"""Default options for iteratively building Click decorators"""


def all_on() -> t.Dict[str, t.Dict]:
    """Return default options with all overrides enabled"""
    options = deepcopy(OPTION_DEFAULTS)
    retval = {}
    # pylint: disable=consider-using-dict-items
    for option in options:
        retval[option] = options[option]
        retval[option]["override"] = OVERRIDE
    return retval


SHOW_EVERYTHING: t.Dict[str, t.Dict] = all_on()
"""Return options for iteratively building Click decorators with all overrides on"""


# Logging schema
def config_logging() -> Schema:
    """
    :returns: A validation schema of all acceptable logging configuration parameter
        names and values with defaults for unset parameters.
    :rtype: :py:class:`~.voluptuous.schema_builder.Schema`

    Logging schema with defaults:

    .. code-block:: yaml

        logging:
          loglevel: INFO
          logfile: None
          logformat: default
          blacklist: ['elastic_transport', 'urllib3']

    """
    return Schema(
        {
            Optional("loglevel", default="INFO"): Any(
                None,
                "NOTSET",
                "DEBUG",
                "INFO",
                "WARNING",
                "ERROR",
                "CRITICAL",
                All(Coerce(int), Any(0, 10, 20, 30, 40, 50)),
            ),
            Optional("logfile", default=None): Any(None, str),
            Optional("logformat", default="default"): Any(
                None, All(Any(str), Any("default", "json", "ecs"))
            ),
            Optional("blacklist", default=["elastic_transport", "urllib3"]): Any(
                None, list
            ),
        }
    )


# All elasticsearch client options, with a few additional arguments.
def config_schema() -> Schema:
    """
    :returns: A validation schema of all acceptable client configuration parameter
        names and values with defaults for unset parameters.
    :rtype: :py:class:`~.voluptuous.schema_builder.Schema`

    The validation schema for an :py:class:`~.elasticsearch9.Elasticsearch` client
    object with defaults
    """
    # pylint: disable=no-value-for-parameter
    return Schema(
        {
            Optional("other_settings", default={}): {
                Optional("master_only", default=False): Boolean(),
                Optional("skip_version_test", default=False): Boolean(),
                Optional("username", default=None): Any(None, str),
                Optional("password", default=None): Any(None, str),
                Optional("api_key", default={}): {
                    Optional("id"): Any(None, str),
                    Optional("api_key"): Any(None, str),
                    Optional("token"): Any(None, str),
                },
            },
            Optional("client", default={}): {
                Optional("hosts", default=None): Any(None, list, str),
                Optional("cloud_id", default=None): Any(None, str),
                Optional("api_key"): Any(None, tuple),
                Optional("basic_auth"): Any(None, tuple),
                Optional("bearer_auth"): Any(None, str),
                Optional("opaque_id"): Any(None, str),
                Optional("headers"): Any(None, dict),
                Optional("connections_per_node"): Any(
                    None, All(Coerce(int), Range(min=1, max=100))
                ),
                Optional("http_compress"): Boolean(),
                Optional("verify_certs"): Boolean(),
                Optional("ca_certs"): Any(None, str),
                Optional("client_cert"): Any(None, str),
                Optional("client_key"): Any(None, str),
                #: Hostname or IP address to verify on the node's certificate.
                #: This is useful if the certificate contains a different value
                #: than the one supplied in ``host``. An example of this situation
                #: is connecting to an IP address instead of a hostname.
                #: Set to ``False`` to disable certificate hostname verification.
                Optional("ssl_assert_hostname"): Any(None, str),
                #: SHA-256 fingerprint of the node's certificate. If this value is
                #: given then root-of-trust verification isn't done and only the
                #: node's certificate fingerprint is verified.
                #:
                #: On CPython 3.10+ this also verifies if any certificate in the
                #: chain including the Root CA matches this fingerprint. However
                #: because this requires using private APIs support for this is
                #: **experimental**.
                Optional("ssl_assert_fingerprint"): Any(None, str),
                Optional("ssl_version"): Any(
                    None, str
                ),  # Minimum acceptable TLS/SSL version
                #: Pre-configured :class:`ssl.SSLContext` OBJECT. If this value
                #: is given then no other TLS options (besides
                #: ``ssl_assert_fingerprint``) can be set on the
                #: :class:`elastic_transport.NodeConfig`.
                Optional("ssl_context"): Any(None, str),
                # Keeping this here in case someone APIs it, but otherwise it's not
                # likely to be used.
                Optional("ssl_show_warn"): Boolean(),
                Optional("transport_class"): Any(None, str),
                Optional("request_timeout"): Any(
                    None, All(Coerce(float), Range(min=0.1, max=86400.0))
                ),
                # node_class: Union[str, Type[BaseNode]] = Urllib3HttpNode,
                Optional("node_class"): Any(None, str),
                # node_pool_class: Type[NodePool] = NodePool,
                Optional("node_pool_class"): Any(None, str),
                Optional("randomize_nodes_in_pool"): Boolean(),
                # node_selector_class: Optional[Union[str, Type[NodeSelector]]] = None,
                Optional("node_selector_class"): Any(None, str),
                Optional("dead_node_backoff_factor"): Any(None, float),
                Optional("max_dead_node_backoff"): Any(None, float),
                # One of:
                # "Serializer"
                # "JsonSerializer"
                # "TextSerializer"
                # "NdjsonSerializer"
                # "CompatibilityModeJsonSerializer"
                # "CompatibilityModeNdjsonSerializer"
                # "MapboxVectorTileSerializer"
                Optional("serializer"): Any(None, str),  # ???
                # :arg serializers: optional dict of serializer instances that will be
                # used for deserializing data coming from the server. (key is the
                # mimetype), e.g.: {'mimetype':'serializer'}
                # "Serializer"
                # "JsonSerializer"
                # "TextSerializer"
                # "NdjsonSerializer"
                # "CompatibilityModeJsonSerializer"
                # "CompatibilityModeNdjsonSerializer"
                # "MapboxVectorTileSerializer"
                Optional("serializers"): Any(None, dict),
                Optional("default_mimetype"): Any(None, str),
                Optional("max_retries"): Any(
                    None, All(Coerce(int), Range(min=1, max=100))
                ),
                # retry_on_status: Collection[int] = (429, 502, 503, 504),
                Optional("retry_on_status"): Any(None, tuple),
                Optional("retry_on_timeout"): Boolean(),
                Optional("sniff_on_start"): Boolean(),
                Optional("sniff_before_requests"): Boolean(),
                Optional("sniff_on_node_failure"): Boolean(),
                Optional("sniff_timeout"): Any(
                    None, All(Coerce(float), Range(min=0.1, max=100.0))
                ),
                Optional("min_delay_between_sniffing"): Any(
                    None, All(Coerce(float), Range(min=1, max=100.0))
                ),
                # Optional[
                #     Callable[
                #         ["Transport", "SniffOptions"],
                #         Union[List[NodeConfig], List[NodeConfig]],
                #     ]
                # ] = None,
                Optional("sniffed_node_callback"): Any(None, str),
                Optional("meta_header"): Boolean(),
                # Cannot specify both 'request_timeout' and 'timeout'
                # Optional('timeout', default=10.0): All(Coerce(float),
                #     Range(min=1, max=120)),
                # Cannot specify both 'randomize_hosts' and 'randomize_nodes_in_pool'
                # Optional('randomize_hosts', default=True): Boolean(),
                Optional("host_info_callback"): Any(
                    None, str
                ),  # ??? needs the name of a callback function
                # Cannot specify both 'sniffer_timeout' and 'min_delay_between_sniffing'
                # Optional('sniffer_timeout', default=0.5): All(Coerce(float),
                #     Range(min=0.1, max=10.0)),
                # Cannot specify both 'sniff_on_connection_fail' and
                #     'sniff_on_node_failure'
                # Optional('sniff_on_connection_fail', default=False): Boolean(),
                # Optional('http_auth'): Any(None, str),
                #     Favor basic_auth instead
                Optional("_transport"): Any(None, str),  # ???
            },
        }
    )


def version_max() -> t.Tuple:
    """Return the max version"""
    return VERSION_MAX


def version_min() -> t.Tuple:
    """Return the min version"""
    return VERSION_MIN


def client_settings() -> t.List[str]:
    """Return the client settings"""
    return CLIENT_SETTINGS


def config_settings() -> t.List[str]:
    """
    Return only the client settings likely to be used in a config file or at the
    command-line.

    This means ignoring some that are valid in
    :py:class:`~.elasticsearch9.Elasticsearch` but are handled different locally.
    Namely, ``api_key`` is handled by :py:class:`~.es_client.builder.OtherArgs`.
    """
    ignore = ["api_key"]
    settings = []
    for setting in CLIENT_SETTINGS:
        if setting not in ignore:
            settings.append(setting)
    return settings


def other_settings() -> t.List[str]:
    """Return the other settings"""
    return OTHER_SETTINGS



"""Test schemacheck module"""

from unittest import TestCase
import pytest
from voluptuous import Schema
from es_client.exceptions import FailedValidation
from es_client.schemacheck import SchemaCheck
from es_client.defaults import (
    config_schema,
    VERSION_MIN,
    version_min,
    VERSION_MAX,
    version_max,
)


class TestSchemaCheck(TestCase):
    """Test SchemaCheck class and member functions"""

    def test_bad_port_value(self):
        """Ensure that a bad port value Raises a FailedValidation"""
        config = {"elasticsearch": {"client": {"port": 70000}}}
        schema = SchemaCheck(config, config_schema(), "elasticsearch", "client")
        with pytest.raises(FailedValidation):
            schema.result()

    def test_entirely_wrong_keys(self):
        """Ensure that unacceptable keys Raises a FailedValidation"""
        config = {
            "elasticsearch": {
                "client_not": {},
                "not_aws": {},
            },
            "something_else": "foo",
        }
        schema = SchemaCheck(config, config_schema(), "elasticsearch", "client")
        with pytest.raises(FailedValidation):
            schema.result()

    def test_does_not_password_filter_non_dict(self):
        """Ensure that if config is not a dictionary that it doesn't choke"""
        config = None
        schema = SchemaCheck(config, Schema(config), "arbitrary", "anylocation")
        assert schema.result() is None


class TestVersionMinMax(TestCase):
    """Test version min and max functions"""

    def test_version_max(self):
        """Ensure version_max returns what it's set with"""
        assert VERSION_MAX == version_max()

    def test_version_min(self):
        """Ensure version_min returns what it's set with"""
        assert VERSION_MIN == version_min()



"""Test config module"""

from unittest import TestCase
import pytest
from dotmap import DotMap  # type: ignore
from elasticsearch9 import Elasticsearch
from es_client.defaults import ES_DEFAULT
from es_client.exceptions import ESClientException
from es_client import config
from . import CACRT, HOST, PASS, USER

CONFIG = {
    "elasticsearch": {
        "other_settings": {"username": USER, "password": PASS},
        "client": {"hosts": HOST, "ca_certs": CACRT},
    }
}


class TestGetClient(TestCase):
    """Test get_client functionality"""

    def test_basic_operation(self):
        """Ensure basic operation"""
        assert isinstance(config.get_client(configdict=CONFIG), Elasticsearch)

    def test_raises_when_no_connection(self):
        """
        Ensures that an exception is raised when it cannot connect to Elasticsearch
        """
        client_args = DotMap()
        other_args = DotMap()
        client_args.update(DotMap(ES_DEFAULT))
        client_args.hosts = ["http://127.0.0.123:12345"]
        client_args.request_timeout = 0.1
        cnf = {
            "elasticsearch": {
                "client": client_args.toDict(),
                "other_settings": other_args.toDict(),
            }
        }
        with pytest.raises(ESClientException):
            _ = config.get_client(configdict=cnf)



"""Test utils module"""

import os
import random
import string
import binascii
from unittest import TestCase
from unittest.mock import Mock
import pytest
from es_client.exceptions import ConfigurationError
from es_client import utils as u
from . import FileTestObj

# pylint: disable=R0903,W0718

DEFAULT = {
    "elasticsearch": {
        "other_settings": {
            "master_only": False,
            "skip_version_test": False,
            "username": None,
            "password": None,
        },
        "client": {
            "hosts": "http://127.0.0.1:9200",
            "request_timeout": 30,
        },
    }
}


# The leading spaces are important here to create a proper yaml file.
YAML = "\n".join(["---", "elasticsearch:", "  client:", "    hosts: {0}"])


def random_envvar(size):
    """Generate a random environment variable"""
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits)
        for _ in range(size)
    )


class TestEnsureList(TestCase):
    """Test the u.ensure_list function"""

    def test_utils_ensure_list_returns_lists(self):
        """
        Test several examples of lists: existing lists, strings, mixed lists/numbers
        """
        verify = ["a", "b", "c", "d"]
        source = ["a", "b", "c", "d"]
        assert verify == u.ensure_list(source)
        verify = ["abcd"]
        source = "abcd"
        assert verify == u.ensure_list(source)
        verify = [["abcd", "defg"], 1, 2, 3]
        source = [["abcd", "defg"], 1, 2, 3]
        assert verify == u.ensure_list(source)
        verify = [{"a": "b", "c": "d"}]
        source = {"a": "b", "c": "d"}
        assert verify == u.ensure_list(source)


class TestPruneNones(TestCase):
    """Test the u.prune_nones function"""

    def test_utils_prune_nones_with(self):
        """Ensure that a dict with a single None value comes back as an empty dict"""
        assert not u.prune_nones({"a": None})

    def test_utils_prune_nones_without(self):
        """Ensure that a dict with no None values comes back unchanged"""
        testval = {"foo": "bar"}
        assert testval == u.prune_nones(testval)


class TestReadFile:
    """Test the u.read_file function"""

    def test_utils_read_file_present(self):
        """Ensure that the written value is what was in the filename"""
        obj = FileTestObj()
        assert obj.written_value == u.read_file(obj.args["filename"])
        obj.teardown()

    def test_raise_when_no_file(self):
        """Raise an exception when there is no file"""
        obj = FileTestObj()
        with pytest.raises(ConfigurationError):
            u.read_file(obj.args["no_file_here"])
        obj.teardown()


class TestReadCerts:
    """Test the u.verify_ssl_paths function"""

    def test_all_as_one(self):
        """Test all 3 possible cert files at once from the same file"""
        obj = FileTestObj()
        config = {
            "ca_certs": obj.args["filename"],
            "client_cert": obj.args["filename"],
            "client_key": obj.args["filename"],
        }
        try:
            u.verify_ssl_paths(config)
        except Exception:
            pytest.fail("Unexpected Exception...")


class TestEnvVars:
    """Test the ability to read environment variables"""

    def test_present(self):
        """Test an existing (present) envvar"""
        obj = FileTestObj()
        evar = random_envvar(8)
        os.environ[evar] = "1234"
        dollar = "${" + evar + "}"
        obj.write_config(obj.args["configfile"], YAML.format(dollar))
        cfg = u.get_yaml(obj.args["configfile"])
        assert cfg["elasticsearch"]["client"]["hosts"] == os.environ.get(evar)
        del os.environ[evar]
        obj.teardown()

    def test_not_present(self):
        """Test a non-existent (not-present) envvar. It should set None here"""
        obj = FileTestObj()
        evar = random_envvar(8)
        dollar = "${" + evar + "}"
        obj.write_config(obj.args["configfile"], YAML.format(dollar))
        cfg = u.get_yaml(obj.args["configfile"])
        assert cfg["elasticsearch"]["client"]["hosts"] is None
        obj.teardown()

    def test_not_present_with_default(self):
        """
        Test a non-existent (not-present) envvar. It should set a default value here
        """
        obj = FileTestObj()
        evar = random_envvar(8)
        default = random_envvar(8)
        dollar = "${" + evar + ":" + default + "}"
        obj.write_config(obj.args["configfile"], YAML.format(dollar))
        cfg = u.get_yaml(obj.args["configfile"])
        assert cfg["elasticsearch"]["client"]["hosts"] == default
        obj.teardown()

    def test_raises_exception(self):
        """Ensure that improper formatting raises a ConfigurationError exception"""
        obj = FileTestObj()
        obj.write_config(
            obj.args["configfile"],
            """
            [weird brackets go here]
            I'm not a yaml file!!!=I have no keys
            I have lots of spaces
            """,
        )
        with pytest.raises(ConfigurationError):
            u.get_yaml(obj.args["configfile"])
        obj.teardown()


class TestVerifyURLSchema:
    """Test the u.verify_url_schema function"""

    def test_full_schema(self):
        """Verify that a proper schema comes back in normalized format"""
        url = "https://127.0.0.1:9200"
        assert u.verify_url_schema(url) == "https://127.0.0.1:9200"

    def test_http_schema_no_port(self):
        """
        Verify that port 80 is added when no port is specified for http
        """
        url = "http://127.0.0.1"
        assert u.verify_url_schema(url) == "http://127.0.0.1:80"

    def test_https_schema_no_port(self):
        """
        Verify that port 443 is added when no port is specified for https
        """
        url = "https://127.0.0.1"
        assert u.verify_url_schema(url) == "https://127.0.0.1:443"

    def test_http_with_path(self):
        """Verify that URLs with paths are accepted and normalized"""
        url = "http://localhost:9200/api/v1"
        assert u.verify_url_schema(url) == "http://localhost:9200"

    def test_https_with_path(self):
        """Verify that HTTPS URLs with paths are accepted and normalized"""
        url = "https://example.com/path/to/resource"
        assert u.verify_url_schema(url) == "https://example.com:443"

    def test_http_no_port_with_path(self):
        """Verify that HTTP URLs without ports but with paths work"""
        url = "http://localhost/api/v1/search"
        assert u.verify_url_schema(url) == "http://localhost:80"

    def test_bad_schema_no_port(self):
        """A URL starting with other than http or https raises an exception w/o port"""
        url = "ftp://127.0.0.1"
        with pytest.raises(ConfigurationError):
            u.verify_url_schema(url)

    def test_bad_schema_with_port(self):
        """A URL starting with other than http or https raises an exception w/port"""
        url = "ftp://127.0.0.1:1234"
        with pytest.raises(ConfigurationError):
            u.verify_url_schema(url)

    def test_invalid_hostname(self):
        """A URL without a valid hostname raises an exception"""
        url = "http://"
        with pytest.raises(ConfigurationError):
            u.verify_url_schema(url)


class TestGetVersion:
    """Test the u.get_version function"""

    def test_positive(self):
        """Ensure that what goes in comes back out unchanged"""
        client = Mock()
        client.info.return_value = {"version": {"number": "9.9.9"}}
        version = u.get_version(client)
        assert version == (9, 9, 9)

    def test_negative(self):
        """Ensure that mismatches are caught"""
        client = Mock()
        client.info.return_value = {"version": {"number": "9.9.9"}}
        version = u.get_version(client)
        assert version != (8, 8, 8)

    def test_dev_version_4_dots(self):
        """Test that anything after a third value and a period is truncated"""
        client = Mock()
        client.info.return_value = {"version": {"number": "9.9.9.dev"}}
        version = u.get_version(client)
        assert version == (9, 9, 9)

    def test_dev_version_with_dash(self):
        """Test that anything after a third value and a dash is truncated"""
        client = Mock()
        client.info.return_value = {"version": {"number": "9.9.9-dev"}}
        version = u.get_version(client)
        assert version == (9, 9, 9)


class TestFileExists:
    """Test the u.file_exists function"""

    def test_positive(self):
        """Ensure that an existing file returns True"""
        obj = FileTestObj()
        obj.write_config(
            obj.args["configfile"],
            """
            [weird brackets go here]
            I'm not a yaml file!!!=I have no keys
            I have lots of spaces
            """,
        )
        assert u.file_exists(obj.args["configfile"])
        obj.teardown()

    def test_negative(self):
        """Ensure that a non-existing file returns False"""
        obj = FileTestObj()
        obj.write_config(
            obj.args["configfile"],
            """
            This file will be deleted
            """,
        )
        obj.teardown()
        assert not u.file_exists(obj.args["configfile"])


class TestParseAPIKeyToken:
    """Test the u.parse_apikey_token function"""

    def success(self):
        """Successfully parse a token"""
        token = "X1VoN0VZY0JJV0lrUTlrdS1QZ2k6QjNZN1VJMlVRd0NHM1VTdHhuNnRKdw=="
        expected = ("_Uh7EYcBIWIkQ9ku-Pgi", "B3Y7UI2UQwCG3UStxn6tJw")
        assert expected == u.parse_apikey_token(token)

    def raises_exception1(self):
        """Raise a binascii.Error when unable to base64 decode a token"""
        token = "Not a valid token"
        with pytest.raises(binascii.Error):
            u.parse_apikey_token(token)

    def raises_exception2(self):
        """Raise an IndexError when able to base64 decode a token, not split by colon"""
        token = "VGhpcyB0ZXh0IGhhcyBubyBjb2xvbg=="
        with pytest.raises(IndexError):
            u.parse_apikey_token(token)



"""
Click commands for es_client CLI

This module defines command-line interface commands for es_client, using
:mod:`click` to handle options and integrate with
:class:`~es_client.builder.Builder`, :mod:`~es_client.logging`,
:mod:`~es_client.config`, and :mod:`~es_client.exceptions`. Commands include
displaying all configuration options, testing Elasticsearch connections, and
testing logging output.

Functions:
    show_all_options: Display all client configuration options and env vars.
    test_connection: Test connection to Elasticsearch.
    test_stderr: Test logging output to stderr.

.. note::
    The ``show_all_options`` command overrides default settings to reveal hidden
    options and show environment variable names, using
    :data:`~es_client.defaults.SHOW_EVERYTHING`.
"""

# pylint: disable=R0913,R0914,R0917,W0613,W0622
import logging
import click
from . import config as cfg
from .defaults import SHOW_EVERYTHING

# SHOW ALL OPTIONS
#
# Below is the ``show-all-options`` command which overrides the default with the values
# in the OVERRIDE constant (``hidden: False`` and ``show_env_vars: True``) which will
# reveal any hidden by default options in the top-level menu so they are exposed in the
# --help output, as well as show the environment variable name that can be used to set
# the option without a flag/argument.
#
# The below options are all included automatically by the decorator:
#
# ``@cfg.options_from_dict(SHOW_EVERYTHING)``
#
# These options require the following other includes:
#
# from es_client.defaults import LOGGING_SETTINGS, ONOFF, OVERRIDE
# from es_client.utils import option_wrapper
# click_opt_wrap = option_wrapper()
#
# @click_opt_wrap(*cli_opts('config', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('hosts', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('cloud_id', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('api_token', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('id', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('api_key', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('username', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('password', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('bearer_auth', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('opaque_id', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('request_timeout', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('http_compress', onoff=ONOFF, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('verify_certs', onoff=ONOFF, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('ca_certs', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('client_cert', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('client_key', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('ssl_assert_hostname', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('ssl_assert_fingerprint', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('ssl_version', override=OVERRIDE))
# @click_opt_wrap(*cli_opts('master-only', onoff=ONOFF, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('skip_version_test', onoff=ONOFF, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('loglevel', settings=LOGGING_SETTINGS, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('logfile', settings=LOGGING_SETTINGS, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('logformat', settings=LOGGING_SETTINGS, override=OVERRIDE))
# @click_opt_wrap(*cli_opts('blacklist', settings=LOGGING_SETTINGS, override=OVERRIDE))

# NOTE: Different procedure for show_all_options than other sub-commands
# Normally, for a sub-command, you would not reset the `cfg.context_settings` as we've
# done here because it also resets the context (ctx). We normally want to pass this
# along from the top level command. In this case, we want it to look like the
# root-level command for the sake of the environment variables being shown for the
# root-level and not a sub-level command.


@click.command(
    context_settings=cfg.context_settings(),
    short_help="Show all client configuration options",
)
@cfg.options_from_dict(SHOW_EVERYTHING)
@click.pass_context
def show_all_options(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
    blacklist,
) -> None:
    """
    Display all client configuration options and environment variables.

    Overrides defaults using :data:`~es_client.defaults.SHOW_EVERYTHING` to reveal
    hidden options and show environment variable names in the --help output.

    Args:
        ctx (:class:`click.Context`): Click context for command execution.
        config (str): Path to YAML configuration file.
        hosts (tuple): Elasticsearch host URLs.
        cloud_id (str): Elastic Cloud ID.
        api_token (str): Base64-encoded API key token.
        id (str): API key ID.
        api_key (str): API key value.
        username (str): Username for basic authentication.
        password (str): Password for basic authentication.
        bearer_auth (str): Bearer authentication token.
        opaque_id (str): Opaque ID for request tracking.
        request_timeout (float): Request timeout in seconds.
        http_compress (bool): Enable HTTP compression.
        verify_certs (bool): Verify SSL certificates.
        ca_certs (str): Path to CA certificates.
        client_cert (str): Path to client certificate.
        client_key (str): Path to client key.
        ssl_assert_hostname (str): SSL hostname to verify.
        ssl_assert_fingerprint (str): SSL certificate fingerprint.
        ssl_version (str): SSL version to use.
        master_only (bool): Connect only to the master node.
        skip_version_test (bool): Skip Elasticsearch version check.
        loglevel (str): Logging level (e.g., DEBUG, INFO).
        logfile (str): Path to log file.
        logformat (str): Log format (e.g., default, json).
        blacklist (tuple): Logger names to exclude.

    Returns:
        None: Outputs help text and exits.

    Example:
        >>> from click import Context, Command
        >>> ctx = Context(Command('show_all_options'), obj={})
        >>> show_all_options(ctx, None, (), None, None, None, None, None, None,
        None, None, None, False, True, None, None, None, None, None, None, False,
        False, None, None, None, ())
        ... # Outputs help text and exits
    """
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    ctx.exit()


@click.command()
@click.pass_context
def test_connection(ctx: click.Context) -> None:
    """
    Test connection to Elasticsearch.

    Uses :func:`~es_client.config.get_client` to create a client from
    :attr:`ctx.obj['configdict'] <click.Context.obj>` and calls
    :meth:`client.info() <elasticsearch9.Elasticsearch.info>` to verify connectivity.

    Args:
        ctx (:class:`click.Context`): Click context with configuration.

    Returns:
        None: Outputs connection result to stdout.

    Raises:
        :exc:`~es_client.exceptions.ESClientException`: If connection fails.

    Example:
        >>> from click import Context, Command
        >>> from unittest.mock import Mock
        >>> ctx = Context(Command('test_connection'), obj={'configdict':
        {'elasticsearch': {'client': {'hosts': ['http://localhost:9200']},
        'other_settings': {}}}})
        >>> client = Mock()
        >>> client.info.return_value = {'version': {'number': '9.0.0'}}
        >>> cfg.get_client = Mock(return_value=client)
        >>> test_connection(ctx)
        Connection result:
        {'version': {'number': '9.0.0'}}
    """
    client = cfg.get_client(configdict=ctx.obj["configdict"])
    click.secho("\nConnection result: ", bold=True)
    click.secho(f"{client.info()}\n")


@click.command()
@click.pass_context
def test_stderr(ctx: click.Context) -> None:
    """
    Test logging output to stderr.

    Logs messages at DEBUG, INFO, WARNING, ERROR, and CRITICAL levels using
    :mod:`logging` to test stderr output configuration.

    Args:
        ctx (:class:`click.Context`): Click context (unused).

    Returns:
        None: Outputs completion message to stdout.

    Example:
        >>> from click import Context, Command
        >>> ctx = Context(Command('test_stderr'), obj={})
        >>> test_stderr(ctx)
        <BLANKLINE>
        Logging test complete.
        <BLANKLINE>
    """
    logger = logging.getLogger(__name__)
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")
    click.secho("\nLogging test complete.\n")


if __name__ == '__main__':
    click.echo("This module is not meant to be run directly.")



"""Test schemacheck module"""

from unittest import TestCase
from copy import deepcopy
import certifi
import click
import pytest
from es_client.builder import Builder
from es_client.exceptions import ConfigurationError
from . import FileTestObj

DEFAULT = {"elasticsearch": {"client": {"hosts": ["http://127.0.0.1:9200"]}}}

YAMLCONFIG = "\n".join(
    ["---", "elasticsearch:", "  client:", "    hosts:", "      - {0}\n"]
)

# pylint: disable=protected-access


def process_cmd(key):
    """Return the key from the click context's object"""
    return click.get_current_context().obj[key]


class TestInit(TestCase):
    """Test initializing a Builder object"""

    def test_read_config_file_old(self):
        """Ensure that the value of es_url is passed to hosts"""
        es_url = "http://127.0.0.1:9200"
        # Build
        file_obj = FileTestObj()
        file_obj.write_config(file_obj.args["configfile"], YAMLCONFIG.format(es_url))
        # Test
        build_obj = Builder(configfile=file_obj.args["configfile"])
        result = list(build_obj.client_args.hosts)
        assert result[0] == es_url
        # Teardown
        file_obj.teardown()

    def test_assign_defaults(self):
        """
        Ensure that the default URL is passed to hosts when an empty config dict is
        passed
        """
        obj = Builder(configdict={})
        assert obj.client_args.hosts == ["http://127.0.0.1:9200"]

    def test_raises_for_both_hosts_and_cloud_id(self):
        """
        Ensure that ConfigurationError is Raised when both hosts and cloud_id are
        passed
        """
        test = {
            "elasticsearch": {
                "client": {"hosts": ["http://10.1.2.3:4567"], "cloud_id": "foo:bar"}
            }
        }
        with pytest.raises(ConfigurationError):
            _ = Builder(configdict=test)

    def test_remove_default_hosts_when_cloud_id(self):
        """
        Ensure that only a default hosts url is removed when cloud_id is also passed
        """
        test = {
            "elasticsearch": {
                "client": {"hosts": ["http://127.0.0.1:9200"], "cloud_id": "foo:bar"}
            }
        }
        obj = Builder(configdict=test)
        assert obj.client_args.hosts is None

    def test_url_schema_validation_fix(self):
        """Ensure that :443 is appended to a host with https and no port"""
        test = {"elasticsearch": {"client": {"hosts": ["https://127.0.0.1"]}}}
        obj = Builder(configdict=test)
        result = list(obj.client_args.hosts)
        assert "https://127.0.0.1:443" == result[0]

    def test_url_schema_validation_raises(self):
        """Ensure that ConfigurationError is raised with an invalid host URL schema"""
        test = {"elasticsearch": {"client": {"hosts": ["127.0.0.1:9200"]}}}
        with pytest.raises(ConfigurationError):
            _ = Builder(configdict=test)


class TestAuth(TestCase):
    """Test authentication methods"""

    def test_user_but_no_pass(self):
        """
        Ensure ConfigurationError is Raised when username is provided but no password
        """
        obj = Builder(configdict=DEFAULT)
        obj.other_args.username = "test"
        assert obj.other_args.password is None
        with pytest.raises(ConfigurationError):
            obj._check_basic_auth()

    def test_pass_but_no_user(self):
        """
        Ensure ConfigurationError is Raised when password is provided but no username
        """
        pwd = "test"
        cfg = deepcopy(DEFAULT)
        cfg["elasticsearch"]["other_settings"] = {}
        cfg["elasticsearch"]["other_settings"]["password"] = pwd
        with pytest.raises(ConfigurationError):
            _ = Builder(configdict=cfg)

    def test_id_but_no_api_key(self):
        """Ensure ConfigurationError is Raised when id is passed but no api_key"""
        test = {
            "elasticsearch": {
                "other_settings": {"api_key": {"id": "test"}},
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        with pytest.raises(ConfigurationError):
            _ = Builder(configdict=test)

    def test_api_key_but_no_id(self):
        """Ensure ConfigurationError is Raised when api_key is passed but no id"""
        test = {
            "elasticsearch": {
                "other_settings": {"api_key": {"api_key": "test"}},
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        with pytest.raises(ConfigurationError):
            _ = Builder(configdict=test)

    def test_no_api_key_values(self):
        """Ensure that API keys remain None"""
        api_id = None
        api_key = None
        api_token = None
        test = {
            "elasticsearch": {
                "other_settings": {
                    "api_key": {"id": api_id, "api_key": api_key, "token": api_token}
                },
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        obj = Builder(configdict=test)
        assert obj.client_args.api_key is None

    def test_proper_api_key(self):
        """Ensure that API key value is assigned to client_args when properly passed"""
        api_id = "foo"
        api_key = "bar"
        test = {
            "elasticsearch": {
                "other_settings": {"api_key": {"id": api_id, "api_key": api_key}},
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        obj = Builder(configdict=test)
        assert tuple(obj._secrets.get_secret('api_key')) == (api_id, api_key)
        assert obj.other_args.api_key.id is None
        assert obj.other_args.api_key.api_key is None
        assert obj.config.other_settings.api_key.id is None
        assert obj.config.other_settings.api_key.api_key is None

    def test_proper_api_key_token(self):
        """Ensure that API key value is assigned to client_args when token is good"""
        api_id = "foo"
        api_key = "bar"
        # token = base64.b64encode(bytes(f'{api_id}:{api_key}', 'utf-8'))
        token = "Zm9vOmJhcg=="
        test = {
            "elasticsearch": {
                "other_settings": {"api_key": {"token": token}},
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        obj = Builder(configdict=test)
        assert tuple(obj._secrets.get_secret('api_key')) == (api_id, api_key)
        assert obj.other_args.api_key.token is None
        assert obj.config.other_settings.api_key.token is None

    def test_invalid_api_key_token(self):
        """Ensure that ConfigurationError is raise when token is invalid"""
        token = "This is an invalid token"
        test = {
            "elasticsearch": {
                "other_settings": {"api_key": {"token": token}},
                "client": {"hosts": ["http://127.0.0.1:9200"]},
            }
        }
        with pytest.raises(ConfigurationError):
            Builder(configdict=test)

    def test_basic_auth_tuple(self):
        """Test basic_auth is set properly"""
        usr = "username"
        pwd = "password"
        cfg = deepcopy(DEFAULT)
        cfg["elasticsearch"]["other_settings"] = {}
        cfg["elasticsearch"]["other_settings"]["username"] = usr
        cfg["elasticsearch"]["other_settings"]["password"] = pwd
        obj = Builder(configdict=cfg)
        obj._check_basic_auth()
        assert usr not in obj.client_args
        assert pwd not in obj.client_args
        assert (usr, pwd) == tuple(obj._secrets.get_secret("basic_auth"))


class TestCheckSSL(TestCase):
    """Ensure that certifi certificates are picked up"""

    def test_certifi(self):
        """
        Ensure that the certifi.where() output matches what was inserted into
        client_args
        """
        https = DEFAULT
        https["elasticsearch"]["client"]["hosts"] = "https://127.0.0.1:9200"
        obj = Builder(configdict=https)
        obj._check_ssl()
        assert certifi.where() == obj.client_args.ca_certs

    def test_ca_certs_named_but_no_file(self):
        """
        Ensure that a ConfigurationError is raised if ca_certs is named but no file
        found
        """
        tmp = FileTestObj()
        tmp.write_config(
            tmp.args["configfile"],
            """
            This file will be deleted
            """,
        )
        tmp.teardown()
        https = {
            "elasticsearch": {
                "client": {
                    "hosts": ["http://127.0.0.1:9200"],
                    "ca_certs": tmp.args["configfile"],
                }
            }
        }
        https["elasticsearch"]["client"]["hosts"] = "https://127.0.0.1:9200"
        with pytest.raises(ConfigurationError):
            Builder(configdict=https)

    # def test_context_for_empty_cloud_id(self):
    #     """Test to see contents of ctx"""
    #     yamlconfig = "\n".join(
    #         [
    #             "---",
    #             "elasticsearch:",
    #             "  client:",
    #             "    hosts:",
    #             "      - 'http://127.0.0.1:9200'",
    #             "    cloud_id: ",
    #         ]
    #     )
    #     # Build
    #     file_obj = FileTestObj()
    #     file_obj.write_config(file_obj.args["configfile"], yamlconfig)
    #     # Test
    #     val = get_yaml(file_obj.args["configfile"])
    #     key = 'draftcfg'
    #     ctx = click.Context(click.Command('cmd'), obj={key: val})
    #     with ctx:
    #         resp = process_cmd(key)
    #     assert resp['logging']['logfile'] is None
    #     # Teardown
    #     file_obj.teardown()



"""Logging configuration for es_client

This module provides utilities for configuring logging in es_client, including filters
(:class:`Whitelist`, :class:`Blacklist`) and a JSON formatter (:class:`JSONFormatter`).
It supports log level conversion, logger name normalization, and setup for CLI and
configuration file inputs, integrating with :py:class:`click.Context`.

Classes:
    Whitelist: Logging filter to allow specific logger names.
    Blacklist: Logging filter to block specific logger names.
    JSONFormatter: Custom formatter for JSON log output.

Functions:
    check_logging_config: Validate logging configuration using SchemaCheck.
    configure_logging: Configure logging from a click context.
    de_dot: Replace dots in logger names with underscores.
    deepmerge: Recursively merge dictionaries for JSON formatting.
    get_format_string: Return a log format string based on log level.
    get_logger: Set up the root logger with handlers and filters.
    get_numeric_loglevel: Convert a string log level to a logging constant.
    override_logging: Merge CLI and config file logging settings.
    check_log_opts: Apply default logging options.
    set_logging: Configure global logging with handlers and filters.
"""

# The __future__ annotations line allows support for Python 3.8 and 3.9
from __future__ import annotations
import typing as t
import sys
import json
import time
import logging
from logging import FileHandler, StreamHandler
from voluptuous import Schema
from click import Context, echo as clicho
import ecs_logging
from .exceptions import LoggingException
from .defaults import config_logging, LOGDEFAULTS
from .schemacheck import SchemaCheck
from .utils import ensure_list, prune_nones

# pylint: disable=R0903

logger = logging.getLogger('')  # Root logger for this module


def check_logging_config(config: t.Dict) -> Schema:
    """
    Validate logging configuration using SchemaCheck.

    Args:
        config (dict): Logging configuration data.

    Returns:
        :class:`voluptuous.Schema`: Validated logging configuration from
            :class:`~es_client.schemacheck.SchemaCheck`.

    Ensures the top-level key ``logging`` is in `config`. Sets an empty default
    dictionary if ``logging`` is absent. Passes the result to
    :class:`~es_client.schemacheck.SchemaCheck` for validation against
    :func:`~es_client.defaults.config_logging`.

    Raises:
        :exc:`TypeError`: If `config` is not a dictionary.

    Example:
        >>> config = {'logging': {'loglevel': 'INFO'}}
        >>> result = check_logging_config(config)
        >>> result['loglevel']
        'INFO'
        >>> config = {}
        >>> result = check_logging_config(config)
        >>> result
        {}
    """
    if not isinstance(config, dict):
        clicho(
            f"Must supply logging information as a dictionary. "
            f'You supplied: "{config}" which is "{type(config)}"'
            f"Using default logging values."
        )
        log_settings = {}
    elif "logging" not in config:
        # None provided. Use defaults.
        log_settings = {}
    else:
        if config["logging"]:
            log_settings = prune_nones(config["logging"])
        else:
            log_settings = {}
    return SchemaCheck(
        log_settings, config_logging(), "Logging Configuration", "logging"
    ).result()


def configure_logging(ctx: Context) -> None:
    """
    Configure logging based on a click context.

    Merges logging settings from :attr:`ctx.obj['draftcfg'] <click.Context.obj>` and
    :attr:`ctx.params <click.Context.params>`, with CLI parameters taking precedence.
    Validates and applies the merged settings using :func:`set_logging`.

    Args:
        ctx (:class:`click.Context`): Click command context containing logging config.

    Example:
        >>> from click import Context, Command
        >>> cfg = {
            'logging': {'loglevel': 'INFO', 'logfile': None, 'logformat': 'default'}
        }
        >>> ctx = Context(Command('cmd'), obj={'draftcfg': cfg}, params={})
        >>> configure_logging(ctx)
        >>> logging.getLogger('').level
        20
    """
    logcfg = override_logging(ctx)
    set_logging(logcfg)


def de_dot(dot_string: str, msg: str) -> t.Dict[str, t.Any]:
    """
    Convert a dotted string and message into a nested dictionary.

    Args:
        dot_string (str): Dotted string (e.g., 'es_client.utils').
        msg (str): Message to nest under the final key.

    Returns:
        dict: Nested dictionary with `msg` as the leaf value.

    Raises:
        :exc:`~es_client.exceptions.LoggingException`: If dictionary creation fails.

    Used by :class:`JSONFormatter` to structure log data.

    Example:
        >>> de_dot('es_client.utils', 'test')
        {'es_client': {'utils': 'test'}}
        >>> de_dot('simple', 'test')
        {'simple': 'test'}
    """
    arr = dot_string.split(".")
    arr.append(msg)
    retval = None
    for idx in range(len(arr), 1, -1):
        if not retval:
            try:
                retval = {arr[idx - 2]: arr[idx - 1]}
            except Exception as err:
                raise LoggingException(err) from err
        else:
            try:
                new_d = {arr[idx - 2]: retval}
                retval = new_d
            except Exception as err:
                raise LoggingException(err) from err
    return retval


def deepmerge(source: t.Dict, destination: t.Dict) -> t.Dict:
    """
    Recursively merge a source dictionary into a destination dictionary.

    Args:
        source (dict): Source dictionary to merge.
        destination (dict): Destination dictionary to update.

    Returns:
        dict: Updated `destination` dictionary.

    Used by :class:`JSONFormatter` to combine log attributes.

    Example:
        >>> source = {'a': {'b': 1}, 'c': 2}
        >>> destination = {'a': {'d': 3}, 'e': 4}
        >>> deepmerge(source, destination)
        {'a': {'b': 1, 'd': 3}, 'e': 4, 'c': 2}
    """
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            deepmerge(value, node)
        else:
            destination[key] = value
    return destination


def get_format_string(nll: int) -> str:
    """
    Return a log format string based on the numeric log level.

    Args:
        nll (int): Numeric log level (e.g., 10 for DEBUG, 20 for INFO).

    Returns:
        str: Format string for :class:`logging.Formatter`.

    Example:
        >>> get_format_string(10)  # DEBUG
        '%(asctime)s %(levelname)-9s %(name)30s %(funcName)23s:%(lineno)-4d %(message)s'
        >>> get_format_string(20)  # INFO
        '%(asctime)s %(levelname)-9s %(message)s'
    """
    return (
        "%(asctime)s %(levelname)-9s %(name)30s "
        "%(funcName)23s:%(lineno)-4d %(message)s"
        if nll == 10
        else "%(asctime)s %(levelname)-9s %(message)s"
    )


def get_logger(log_opts: t.Dict) -> None:
    """
    Configure the root logger with appropriate handlers.

    If `logfile` is provided in `log_opts`, uses a :class:`logging.FileHandler`.
    Otherwise, splits logs into :class:`logging.StreamHandler` instances for stdout
    (up to INFO) and stderr (WARNING and above). Applies formatters and filters based
    on `logformat` and `blacklist`.

    Args:
        log_opts (dict): Logging configuration with keys: loglevel, logfile, logformat,
            blacklist.

    Raises:
        OSError: If `logfile` cannot be opened.
        ValueError: If `loglevel` is invalid.

    Example:
        >>> log_opts = {
            'loglevel': 'INFO',
            'logfile': None,
            'logformat': 'default',
            'blacklist': []
        }
        >>> get_logger(log_opts)
        >>> len(logging.getLogger('').handlers) >= 1
        True
    """
    logfile = log_opts.get("logfile", None)
    kind = log_opts.get("logformat", "default")
    nll = get_numeric_loglevel(log_opts.get("loglevel", "INFO"))

    logger.setLevel(nll)

    handler_map = {
        "logfile": FileHandler(logfile) if logfile else None,
        "stdout": StreamHandler(stream=sys.stdout),
        "stderr": StreamHandler(stream=sys.stderr),
    }
    format_map = {
        "default": logging.Formatter(get_format_string(nll)),
        "json": JSONFormatter(),
        "ecs": ecs_logging.StdlibFormatter(),
    }

    def add_handler(source: t.Literal['logfile', 'stdout', 'stderr']) -> None:
        handler = handler_map[source]
        handler.setFormatter(format_map[kind])
        handler.setLevel(nll)
        if source == 'stdout':
            handler.addFilter(lambda record: record.levelno <= logging.INFO)
        if source == 'stderr':
            handler.setLevel(logging.WARNING)
            fltr = max(logging.WARNING, nll)
            handler.addFilter(lambda record: record.levelno >= fltr)
            for entry in ensure_list(log_opts["blacklist"]):
                handler.addFilter(Blacklist(entry))

        logger.addHandler(handler)

    if logfile:
        add_handler('logfile')
    else:
        add_handler('stdout')
        add_handler('stderr')


def get_numeric_loglevel(level: str) -> int:
    """
    Convert a string log level to a logging module constant.

    Args:
        level (str): Log level name (e.g., 'DEBUG', 'INFO').

    Returns:
        int: Corresponding logging constant (e.g., :data:`logging.DEBUG`).

    Raises:
        ValueError: If the level is not a valid log level.

    The mapping is:

    .. list-table:: Log Levels
       :widths: 10 5 85
       :header-rows: 1

       * - Level
         - #
         - Description
       * - NOTSET
         - 0
         - When set on a logger, ancestor loggers determine the effective level.
           If NOTSET, all events are logged. On a handler, all events are handled.
       * - DEBUG
         - 10
         - Detailed information for diagnosing problems.
       * - INFO
         - 20
         - Confirmation that things are working as expected.
       * - WARNING
         - 30
         - Indicates an unexpected issue or potential problem (e.g., 'disk space low').
       * - ERROR
         - 40
         - A serious problem preventing some functionality.
       * - CRITICAL
         - 50
         - A severe error, possibly halting the program.

    Example:
        >>> get_numeric_loglevel('DEBUG')
        10
        >>> get_numeric_loglevel('INFO')
        20
        >>> get_numeric_loglevel('INVALID')
        Traceback (most recent call last):
            ...
        ValueError: Invalid log level: INVALID
    """
    numeric_log_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_log_level, int):
        raise ValueError(f"Invalid log level: {level}")
    return numeric_log_level


def override_logging(ctx: Context) -> t.Dict:
    """
    Merge CLI and config file logging settings.

    Retrieves logging configuration from :attr:`ctx.obj['draftcfg'] <click.Context.obj>`
    and overrides with :attr:`ctx.params <click.Context.params>`, with CLI taking
    precedence. Validates the config using :func:`check_logging_config`.

    Args:
        ctx (:class:`click.Context`): Click command context.

    Returns:
        dict: Merged and validated logging configuration.

    Example:
        >>> from click import Context, Command
        >>> cfg = {'logging': {'loglevel': 'INFO'}}
        >>> ctx = Context(
            Command('cmd'), obj={'draftcfg': cfg}, params={'loglevel': 'DEBUG'}
        )
        >>> result = override_logging(ctx)
        >>> result['loglevel']
        'DEBUG'
    """
    init_logcfg = check_logging_config(ctx.obj["draftcfg"])
    debug = "loglevel" in init_logcfg and init_logcfg["loglevel"] == "DEBUG"
    if "loglevel" in ctx.params and ctx.params["loglevel"] is not None:
        debug = ctx.params["loglevel"] == "DEBUG"

    paramlist = ["loglevel", "logfile", "logformat", "blacklist"]
    for entry in paramlist:
        if entry in ctx.params:
            if not ctx.params[entry]:
                continue
            if (
                debug
                and init_logcfg[entry] is not None
                and init_logcfg["loglevel"] != "DEBUG"
            ):
                clicho(
                    f"DEBUG: Overriding configuration file setting {entry}="
                    f"{init_logcfg[entry]} with command-line option {entry}="
                    f"{ctx.params[entry]}"
                )
            if entry == "blacklist":
                init_logcfg[entry] = list(ctx.params[entry])
            else:
                init_logcfg[entry] = ctx.params[entry]
    return init_logcfg


def check_log_opts(log_opts: t.Dict) -> t.Dict:
    """
    Apply default logging options to unset keys.

    Args:
        log_opts (dict): Logging configuration data.

    Returns:
        dict: Updated `log_opts` with defaults from
            :data:`~es_client.defaults.LOGDEFAULTS`.

    Example:
        >>> log_opts = {'loglevel': 'INFO'}
        >>> result = check_log_opts(log_opts)
        >>> result['loglevel']
        'INFO'
        >>> 'logfile' in result
        True
    """
    for k, v in LOGDEFAULTS.items():
        log_opts[k] = v if k not in log_opts else log_opts[k]
    return log_opts


def set_logging(options: t.Dict) -> None:
    """
    Configure global logging options.

    Applies settings from `options` to the root logger, attaching handlers and filters.
    Adds a :class:`logging.NullHandler` for the 'elasticsearch9.trace' logger
    to suppress trace logs.

    Args:
        options (dict): Logging configuration with keys: loglevel, logfile, logformat,
            blacklist.

    Raises:
        OSError: If `logfile` cannot be opened.
        ValueError: If `loglevel` is invalid.

    Example:
        >>> options = {
            'loglevel': 'INFO',
            'logfile': None,
            'logformat': 'default',
            'blacklist': []
        }
        >>> set_logging(options)
        >>> len(logging.getLogger('').handlers) >= 1
        True
    """
    log_opts = check_log_opts(options)
    get_logger(log_opts)
    logging.getLogger("elasticsearch9.trace").addHandler(logging.NullHandler())
    if log_opts["blacklist"]:
        for entry in ensure_list(log_opts["blacklist"]):
            for handler in logging.root.handlers:
                handler.addFilter(Blacklist(entry))


class Whitelist(logging.Filter):
    """
    Logging filter to allow only specified logger names.

    Args:
        *whitelist (list): Logger names to allow (e.g., ['es_client']).

    Attributes:
        whitelist (list): List of :class:`logging.Filter` objects for allowed names.

    Example:
        >>> import logging
        >>> record = logging.makeLogRecord({'name': 'es_client.test', 'msg': 'Test'})
        >>> whitelist = Whitelist(['es_client'])
        >>> whitelist.filter(record)
        True
        >>> record = logging.makeLogRecord({'name': 'other.test', 'msg': 'Test'})
        >>> whitelist.filter(record)
        False
    """

    def __init__(self, *whitelist: t.List[str]):
        super().__init__()
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter log records based on logger name."""
        return any(f.filter(record) for f in self.whitelist)

    def __repr__(self) -> str:
        """Return a string representation of the filter."""
        return f"<Whitelist names={[f.name for f in self.whitelist]}>"


class Blacklist(Whitelist):
    """
    Logging filter to block specified logger names.

    Inherits from :class:`Whitelist`, inverting the filter logic to block listed names.

    Args:
        *blacklist (list): Logger names to block (e.g., ['es_client.utils']).

    Attributes:
        whitelist (list): List of :class:`logging.Filter` objects for blocked names.

    Example:
        >>> import logging
        >>> record = logging.makeLogRecord({'name': 'es_client.test', 'msg': 'Test'})
        >>> blacklist = Blacklist(['es_client.test'])
        >>> blacklist.filter(record)
        False
        >>> record = logging.makeLogRecord({'name': 'other.test', 'msg': 'Test'})
        >>> blacklist.filter(record)
        True
    """

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter log records based on logger name."""
        return not super().filter(record)

    def __repr__(self) -> str:
        """Return a string representation of the filter."""
        return f"<Blacklist names={[f.name for f in self.whitelist]}>"


class JSONFormatter(logging.Formatter):
    """
    Custom logging formatter for JSON output.

    Formats log records as JSON objects with timestamp, log level, logger name,
    function, line number, and message.

    Example:
        >>> import logging
        >>> record = logging.makeLogRecord({
        ...     'name': 'test', 'levelname': 'INFO', 'msg': 'Test message',
        ...     'funcName': 'test_func', 'lineno': 42, 'created': 1625097600.0,
        ...     'msecs': 123.456
        ... })
        >>> formatter = JSONFormatter()
        >>> formatted = formatter.format(record)
        >>> 'test' in formatted and 'INFO' in formatted
        True
    """

    WANTED_ATTRS = {
        "levelname": "loglevel",
        "funcName": "function",
        "lineno": "linenum",
        "message": "message",
        "name": "name",
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format a log record as a JSON string."""
        self.converter = time.gmtime
        fmt = "%Y-%m-%dT%H:%M:%S"
        mil = str(record.msecs).split(".", maxsplit=1)[0]
        timestamp = f"{self.formatTime(record, datefmt=fmt)}.{mil}Z"
        result = {"@timestamp": timestamp}
        available = record.__dict__
        available["message"] = record.getMessage()
        for attribute in set(self.WANTED_ATTRS).intersection(available):
            result = deepmerge(
                de_dot(self.WANTED_ATTRS[attribute], getattr(record, attribute)), result
            )
        if "message" not in result:
            result["message"] = available["message"]
        return json.dumps(result, sort_keys=True)

    def __repr__(self) -> str:
        """Return a string representation of the formatter."""
        return "<JSONFormatter>"



"""
Sample CLI script that will get a client using both config file and CLI args/options
"""

# pylint: disable=E1120,R0917
import click
from elasticsearch9.exceptions import BadRequestError, NotFoundError
from es_client import config as cfg
from es_client.defaults import OPTION_DEFAULTS
from es_client.logging import configure_logging
from es_client.commands import show_all_options, test_connection, test_stderr

# Since this file will not be imported, we use this to squash the F401 error
__all__ = ["BadRequestError", "NotFoundError"]


# The following default options are all automatically added by the decorator:
#
# ``@cfg.options_from_dict(OPTION_DEFAULTS)``
#
# Be sure to add any other options or arguments either before or after this decorator,
# and add any added arguments in ``def run()``, preserving their order in both
# locations.  ``ctx`` needs to be the first arg after ``def run()`` as a special
# argument for Click, and does not need a decorator function.

# These options require the following other includes:
#
# from es_client.defaults import LOGGING_SETTINGS, ONOFF
# from es_client.utils import option_wrapper
# click_opt_wrap = option_wrapper()
#
# @click_opt_wrap(*cli_opts('config'))
# @click_opt_wrap(*cli_opts('hosts'))
# @click_opt_wrap(*cli_opts('cloud_id'))
# @click_opt_wrap(*cli_opts('api_token'))
# @click_opt_wrap(*cli_opts('id'))
# @click_opt_wrap(*cli_opts('api_key'))
# @click_opt_wrap(*cli_opts('username'))
# @click_opt_wrap(*cli_opts('password'))
# @click_opt_wrap(*cli_opts('bearer_auth'))
# @click_opt_wrap(*cli_opts('opaque_id'))
# @click_opt_wrap(*cli_opts('request_timeout'))
# @click_opt_wrap(*cli_opts('http_compress', onoff=ONOFF))
# @click_opt_wrap(*cli_opts('verify_certs', onoff=ONOFF))
# @click_opt_wrap(*cli_opts('ca_certs'))
# @click_opt_wrap(*cli_opts('client_cert'))
# @click_opt_wrap(*cli_opts('client_key'))
# @click_opt_wrap(*cli_opts('ssl_assert_hostname'))
# @click_opt_wrap(*cli_opts('ssl_assert_fingerprint'))
# @click_opt_wrap(*cli_opts('ssl_version'))
# @click_opt_wrap(*cli_opts('master-only', onoff=ONOFF))
# @click_opt_wrap(*cli_opts('skip_version_test', onoff=ONOFF))
# @click_opt_wrap(*cli_opts('loglevel', settings=LOGGING_SETTINGS))
# @click_opt_wrap(*cli_opts('logfile', settings=LOGGING_SETTINGS))
# @click_opt_wrap(*cli_opts('logformat', settings=LOGGING_SETTINGS))
# @click_opt_wrap(*cli_opts('blacklist', settings=LOGGING_SETTINGS))


# pylint: disable=R0913,R0914,W0613,W0622


@click.group(context_settings=cfg.context_settings())
@cfg.options_from_dict(OPTION_DEFAULTS)
@click.version_option(None, "-v", "--version", prog_name="cli_example")
@click.pass_context
def run(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
    blacklist,
):
    """
    CLI Example

    Any text added to a docstring will show up in the --help/usage output.

    Set short_help='' in @func.command() definitions for each command for terse
    descriptions in the main help/usage output, as with show_all_options() in this
    example.
    """
    # If there's a default file location for client configuration, e.g.
    # $HOME/.curator/curator.yml, then specify it here. ctx.obj is now instantiated in
    # ``config.cfg.context_settings()``
    ctx.obj["default_config"] = None

    # The ``cfg.get_config`` function will grab the configuration derived from a YAML
    # config file specified in command-line parameters, or if that is unspecified but
    # ctx.obj['default_config'] is provided, use that. If quiet=True, suppress the line
    # written to STDOUT that indicates the file at ctx.obj['default_config'] is being
    # used.
    # If neither ctx.params['config'] nor ctx.obj['default_config'] reference a YAML
    # configuration file, then a config dict with empty/default configured is
    # generated. The result is stored in ctx.obj['draftcfg']
    cfg.get_config(ctx, quiet=False)

    # Configure logging. This will use the values from command line parameters, or
    # what's now been stored in ctx.obj['draftcfg']
    configure_logging(ctx)

    # The ``cfg.generate_configdict`` function does all of the overriding of YAML
    # config file options by command-line specified ones and stores the ready-to-be-
    # used by Builder configuration in ctx.obj['configdict']
    cfg.generate_configdict(ctx)


run.add_command(show_all_options)
run.add_command(test_connection)
run.add_command(test_stderr)

if __name__ == "__main__":
    run()



"""Test cli_example"""

from os import devnull
from unittest import TestCase
from click import testing as clicktest
from es_client.cli_example import run
from . import CACRT, HOST, PASS, USER
from ..unit import FileTestObj

YAMLCONFIG = "\n".join(
    [
        "---",
        "logging:",
        "  loglevel: INFO",
        "  logfile:",
        "  logformat: default",
        "  blacklist: ['elastic_transport', 'urllib3']",
    ]
)


class TestCLIExample(TestCase):
    """Test CLI Example"""

    def test_basic_operation(self):
        "Ensure basic functionality"
        args = [
            "--hosts",
            HOST,
            "--username",
            USER,
            "--password",
            PASS,
            "--ca_certs",
            CACRT,
            "--loglevel",
            "DEBUG",
            "test-connection",
        ]
        runner = clicktest.CliRunner()
        result = runner.invoke(run, args)
        assert result.exit_code == 0

    def test_show_all_options(self):
        """Ensure show-all-options works"""
        args = ["show-all-options"]
        runner = clicktest.CliRunner()
        result = runner.invoke(run, args=args)
        assert result.exit_code == 0

    def test_logging_options_json(self):
        """Testing JSON log options"""
        args = [
            "--hosts",
            HOST,
            "--username",
            USER,
            "--password",
            PASS,
            "--ca_certs",
            CACRT,
            "--loglevel",
            "DEBUG",
            "--logformat",
            "json",
            "test-connection",
        ]
        runner = clicktest.CliRunner()
        result = runner.invoke(run, args=args)
        assert result.exit_code == 0

    def test_logging_options_ecs(self):
        """Testing ECS log options"""
        args = [
            "--hosts",
            HOST,
            "--username",
            USER,
            "--password",
            PASS,
            "--ca_certs",
            CACRT,
            "--loglevel",
            "WARNING",
            "--logfile",
            devnull,
            "--logformat",
            "ecs",
            "test-connection",
        ]
        runner = clicktest.CliRunner()
        result = runner.invoke(run, args)
        assert result.exit_code == 0

    def test_logging_options_from_config_file(self):
        """Testing logging options from a config file"""
        # Build
        file_obj = FileTestObj()
        file_obj.write_config(file_obj.args["configfile"], YAMLCONFIG)
        # Test
        args = [
            "--config",
            file_obj.args["configfile"],
            "--hosts",
            HOST,
            "--username",
            USER,
            "--password",
            PASS,
            "--ca_certs",
            CACRT,
            "test-connection",
        ]
        runner = clicktest.CliRunner()
        result = runner.invoke(run, args)
        assert result.exit_code == 0
        # Teardown
        file_obj.teardown()



"""Schema validation and redaction utilities

This module provides the :class:`SchemaCheck` class to validate configuration
dictionaries against a :class:`voluptuous.Schema` and a function to redact sensitive
data for secure logging. It supports configuration validation for
:class:`~es_client.builder.Builder` and logging in :mod:`~es_client.logging`.

Classes:
    SchemaCheck: Validates a configuration dictionary against a schema.

Functions:
    password_filter: Redact sensitive values from a configuration dictionary.
"""

# pylint: disable=E1101,W0718
import typing as t
import logging
from re import sub
from copy import deepcopy
from voluptuous import Schema
from .debug import debug, begin_end
from .defaults import KEYS_TO_REDACT
from .exceptions import FailedValidation

logger = logging.getLogger(__name__)


def password_filter(data: t.Dict) -> t.Dict:
    """
    Redact sensitive values from a configuration dictionary.

    Args:
        data (dict): Configuration dictionary to process.

    Returns:
        dict: A deep copy of `data` with sensitive values (keys in
            :data:`~es_client.defaults.KEYS_TO_REDACT`) replaced with 'REDACTED'.

    Recursively traverses `data`, replacing values of keys listed in
    :data:`~es_client.defaults.KEYS_TO_REDACT` (e.g., 'password', 'api_key') with
    'REDACTED' for secure logging.

    Example:
        >>> data = {'user': 'test', 'password': 'secret', 'nested': {'api_key': 'key'}}
        >>> filtered = password_filter(data)
        >>> filtered
        {'user': 'test', 'password': 'REDACTED', 'nested': {'api_key': 'REDACTED'}}
        >>> data['password']  # Original unchanged
        'secret'
    """

    def iterdict(mydict):
        for key, value in mydict.items():
            if isinstance(value, dict):
                iterdict(value)
            elif key in KEYS_TO_REDACT:
                mydict.update({key: "REDACTED"})
        return mydict

    return iterdict(deepcopy(data))


class SchemaCheck:
    """
    Validate a configuration dictionary against a voluptuous schema.

    Args:
        config (dict): Configuration dictionary to validate.
        schema (:class:`voluptuous.Schema`): Schema to validate against.
        test_what (str): Description of the configuration block (e.g., 'Client Config').
        location (str): Context of the configuration (e.g., 'elasticsearch.client').

    Attributes:
        config (dict): The configuration dictionary.
        schema (:class:`voluptuous.Schema`): The validation schema.
        test_what (str): Description of the configuration block.
        location (str): Context of the configuration.
        badvalue (str): Invalid value causing validation failure, or 'no bad value yet'.
        error (str): Validation error message, or 'No error yet'.

    Raises:
        :exc:`~es_client.exceptions.FailedValidation`: If validation fails.

    Example:
        >>> from voluptuous import Schema
        >>> schema = Schema({'host': str})
        >>> config = {'host': 'localhost'}
        >>> check = SchemaCheck(config, schema, 'Test Config', 'test')
        >>> check.result()
        {'host': 'localhost'}
        >>> config = {'host': 123}
        >>> check = SchemaCheck(config, schema, 'Test Config', 'test')
        >>> check.result()
        Traceback (most recent call last):
            ...
        es_client.exceptions.FailedValidation: Configuration: Test Config: Location:
        test: Bad Value: "123", expected str @ data['host']. Check configuration file.
    """

    def __init__(self, config: t.Dict, schema: Schema, test_what: str, location: str):
        debug.lv2('Starting function...')
        debug.lv5(f'Schema: {schema}')
        if isinstance(config, dict):
            debug.lv5(f'"{test_what} config: {password_filter(config)}"')
        else:
            debug.lv5(f'"{test_what} config: {config}"')
        self.config = config
        self.schema = schema
        self.test_what = test_what
        self.location = location
        self.badvalue = "no bad value yet"
        self.error = "No error yet"

    @begin_end()
    def parse_error(self) -> t.Any:
        """
        Extract and report the invalid value causing a validation error.

        Attempts to parse the error message to identify the bad value, updating
        :attr:`badvalue`. Logs errors if parsing fails.

        Returns:
            None: Updates :attr:`badvalue` and logs the result.

        Example:
            >>> from voluptuous import Schema
            >>> schema = Schema({'host': str})
            >>> config = {'host': 123}
            >>> check = SchemaCheck(config, schema, 'Test Config', 'test')
            >>> try:
            ...     check.result()
            ... except FailedValidation:
            ...     check.badvalue
            '123'
        """

        def get_badvalue(data_string, data):
            debug.lv5('Starting nested function...')
            elements = sub(r"[\'\]]", "", data_string).split("[")
            elements.pop(0)  # Remove 'data' prefix
            value = None
            for k in elements:
                try:
                    debug.lv4('TRY: parsing key')
                    key = int(k)
                except ValueError:
                    key = k
                if value is None:
                    value = data[key]
            debug.lv5(f'Exiting nested function, returning {value}')
            return value

        try:
            debug.lv4('TRY: parsing error')
            self.badvalue = get_badvalue(str(self.error).split()[-1], self.config)
        except Exception as exc:
            logger.error(f'Unable to extract value: {exc}')
            self.badvalue = "(could not determine)"

    @begin_end()
    def result(self) -> t.Any:
        """
        Validate the configuration and return the result.

        Returns:
            :class:`voluptuous.Schema`: Validated configuration from
                :attr:`config` if successful.

        Raises:
            :exc:`~es_client.exceptions.FailedValidation`: If validation fails,
                including error details and bad value.

        Calls :meth:`parse_error` to extract the invalid value if validation fails.

        Example:
            >>> from voluptuous import Schema
            >>> schema = Schema({'host': str})
            >>> config = {'host': 'localhost'}
            >>> check = SchemaCheck(config, schema, 'Test Config', 'test')
            >>> check.result()
            {'host': 'localhost'}
        """
        try:
            debug.lv4('TRY: validating configuration...')
            return self.schema(self.config)
        except Exception as exc:
            try:
                debug.lv4('TRY: parsing exception...')
                self.error = exc.errors[0]  # type: ignore[attr-defined]
            except Exception as err:
                logger.error(f'Could not parse exception: {err}')
                self.error = f"{exc}"
            self.parse_error()
            logger.error(f'Schema error: {self.error}')
            msg = (
                f"Configuration: {self.test_what}: Location: {self.location}: "
                f'Bad Value: "{self.badvalue}", {self.error}. Check configuration file.'
            )
            debug.lv3('Exiting function, raising exception')
            debug.lv5(f'Value = "{exc}"')
            logger.error(msg)
            raise FailedValidation(msg) from exc

    def __repr__(self) -> str:
        """
        Return a string representation of the SchemaCheck instance.

        Returns:
            str: Description of the configuration being validated.

        Example:
            >>> from voluptuous import Schema
            >>> check = SchemaCheck(
                {'host': 'localhost'}, Schema({'host': str}), 'Test Config', 'test'
            )
            >>> repr(check)
            "<SchemaCheck test_what='Test Config' location='test'>"
        """
        return f"<SchemaCheck test_what='{self.test_what}' location='{self.location}'>"



"""Test logging module"""

from io import StringIO
import logging
import json
import re
import tempfile
import unittest
from unittest.mock import MagicMock
from unittest import TestCase
import pytest
import click
from es_client.utils import get_yaml
from es_client.logging import (
    Whitelist,
    Blacklist,
    JSONFormatter,
    check_logging_config,
    override_logging,
    get_logger,
    get_numeric_loglevel,
    get_format_string,
    check_log_opts,
    de_dot,
    deepmerge,
)
from . import FileTestObj


def process_cmd(key):
    """Return the key from the click context's object"""
    return click.get_current_context().obj[key]


class TestCheckLoggingConfig(TestCase):
    """Test check_logging_config functionality"""

    default = {
        "loglevel": "INFO",
        "blacklist": ["elastic_transport", "urllib3"],
        "logfile": None,
        "logformat": "default",
    }

    def test_non_dict(self):
        """Ensure it yields default values"""
        assert self.default == check_logging_config("not-a-dict")

    def test_empty_key(self):
        """Ensure it yields default values too"""
        assert self.default == check_logging_config({"logging": {}})

    def test_logging_context_for_empty_logfile(self):
        """Test to see contents of ctx"""
        yamlconfig = "\n".join(
            [
                "---",
                "logging:",
                "  loglevel: INFO",
                "  logfile: ",
                "  logformat: default",
                "  blacklist: ['elastic_transport', 'urllib3']",
            ]
        )
        # Build
        file_obj = FileTestObj()
        file_obj.write_config(file_obj.args["configfile"], yamlconfig)
        # Test
        val = get_yaml(file_obj.args["configfile"])
        key = 'draftcfg'
        ctx = click.Context(click.Command('cmd'), obj={key: val})
        with ctx:
            resp = process_cmd(key)
        assert resp['logging']['logfile'] is None
        # Teardown
        file_obj.teardown()


class TestGetNumericLogLevel(TestCase):
    """Test get_numeric_loglevel function"""

    def test_invalid_loglevel(self):
        """Ensure it raises an exception when an invalid loglevel is provided"""
        with pytest.raises(ValueError):
            get_numeric_loglevel("NONSENSE")


# Test custom logging filters
class TestLoggingFilters(unittest.TestCase):
    """Test custom logging filters."""

    def test_whitelist_filter(self):
        """Test that Whitelist filter allows only specified logger names."""
        whitelist = Whitelist('test_logger')
        record_allowed = logging.LogRecord(
            'test_logger', logging.INFO, 'path', 1, 'message', None, None
        )
        record_blocked = logging.LogRecord(
            'other_logger', logging.INFO, 'path', 1, 'message', None, None
        )
        self.assertTrue(whitelist.filter(record_allowed))
        self.assertFalse(whitelist.filter(record_blocked))

    def test_blacklist_filter(self):
        """Test that Blacklist filter blocks specified logger names."""
        blacklist = Blacklist('test_logger')
        record_blocked = logging.LogRecord(
            'test_logger', logging.INFO, 'path', 1, 'message', None, None
        )
        record_allowed = logging.LogRecord(
            'other_logger', logging.INFO, 'path', 1, 'message', None, None
        )
        self.assertFalse(blacklist.filter(record_blocked))
        self.assertTrue(blacklist.filter(record_allowed))


# Test JSONFormatter
class TestJSONFormatter(unittest.TestCase):
    """Test JSONFormatter class."""

    def test_format(self):
        """Test that JSONFormatter correctly formats log records into JSON."""
        formatter = JSONFormatter()
        record = logging.LogRecord(
            'test_logger', logging.INFO, 'path', 1, 'Test message', None, None
        )
        formatted = formatter.format(record)
        data = json.loads(formatted)
        self.assertIn('@timestamp', data)
        self.assertIn('message', data)
        self.assertEqual(data['message'], 'Test message')
        self.assertIn('loglevel', data)
        self.assertEqual(data['loglevel'], 'INFO')
        # Verify timestamp format (ISO 8601 with milliseconds)
        self.assertTrue(
            re.match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z', data['@timestamp'])
        )


# Test configuration-related functions
class TestConfigurationFunctions(unittest.TestCase):
    """Test configuration-related functions."""

    def test_check_logging_config_valid(self):
        """Test check_logging_config with a valid configuration."""
        valid_config = {'logging': {'loglevel': 'DEBUG', 'logformat': 'json'}}
        result = check_logging_config(valid_config)
        self.assertEqual(result['loglevel'], 'DEBUG')
        self.assertEqual(result['logformat'], 'json')

    def test_check_logging_config_invalid(self):
        """Test check_logging_config with an invalid log level."""
        invalid_config = {'logging': {'loglevel': 'INVALID'}}
        with self.assertRaises(
            Exception
        ):  # Schema validation should raise an exception
            check_logging_config(invalid_config)

    def test_check_logging_config_no_config(self):
        """Test check_logging_config with no configuration provided."""
        no_config = {}
        result = check_logging_config(no_config)
        self.assertEqual(result['loglevel'], 'INFO')  # Default value

    def test_override_logging(self):
        """Test override_logging merges CLI options over config file settings."""
        ctx = MagicMock()
        ctx.obj = {'draftcfg': {'logging': {'loglevel': 'INFO'}}}
        ctx.params = {'loglevel': 'DEBUG'}
        result = override_logging(ctx)
        self.assertEqual(result['loglevel'], 'DEBUG')


# Test logger setup and related functions
class TestLoggerSetup(unittest.TestCase):
    """Test logger setup and related functions."""

    def setUp(self):
        """Reset logging configuration before each test."""
        logging.root.handlers = []
        logging.root.setLevel(logging.NOTSET)

    def tearDown(self):
        """Reset logging configuration after each test."""
        logging.root.handlers = []
        logging.root.setLevel(logging.NOTSET)

    def test_get_logger_with_logfile(self):
        """Test get_logger with a logfile specified."""
        with tempfile.NamedTemporaryFile() as tmpfile:
            log_opts = {
                'loglevel': 'INFO',
                'logfile': tmpfile.name,
                'logformat': 'default',
                'blacklist': [],
            }
            get_logger(log_opts)
            logger = logging.getLogger('test_logger_with_logfile')
            logger.info('Test message')
            with open(tmpfile.name, 'r', encoding='utf8') as f:
                content = f.read()
                self.assertIn('Test message', content)

    def test_get_logger_without_logfile(self):
        """Test get_logger without a logfile, using stream handlers."""
        log_opts = {'loglevel': 'INFO', 'logformat': 'default', 'blacklist': []}
        get_logger(log_opts)
        logger = logging.getLogger('test_logger_without_logfile')
        stdout = StringIO()
        stderr = StringIO()
        handler_stdout = logging.StreamHandler(stdout)
        handler_stderr = logging.StreamHandler(stderr)
        logger.addHandler(handler_stdout)
        logger.addHandler(handler_stderr)
        logger.info('Test info message')
        logger.error('Test error message')
        self.assertIn('Test info message', stdout.getvalue())
        self.assertIn('Test error message', stderr.getvalue())

    def test_get_numeric_loglevel(self):
        """Test conversion of string log levels to numeric values."""
        self.assertEqual(get_numeric_loglevel('DEBUG'), 10)
        self.assertEqual(get_numeric_loglevel('INFO'), 20)
        with self.assertRaises(ValueError):
            get_numeric_loglevel('INVALID')


# Test utility functions
class TestUtilityFunctions(unittest.TestCase):
    """Test utility functions."""

    def test_get_format_string(self):
        """Test format string selection based on log level."""
        debug_format = get_format_string(10)  # DEBUG
        info_format = get_format_string(20)  # INFO
        self.assertIn('%(funcName)', debug_format)  # Detailed format for DEBUG
        self.assertNotIn('%(funcName)', info_format)  # Simpler format for INFO

    def test_check_log_opts(self):
        """Test that check_log_opts applies defaults to partial configs."""
        partial_opts = {'loglevel': 'DEBUG'}
        result = check_log_opts(partial_opts)
        self.assertEqual(result['loglevel'], 'DEBUG')
        self.assertEqual(result['logfile'], None)  # Default
        self.assertEqual(result['logformat'], 'default')  # Default

    def test_de_dot(self):
        """Test conversion of dotted strings to nested dictionaries."""
        self.assertEqual(de_dot('loglevel', 'INFO'), {'loglevel': 'INFO'})
        self.assertEqual(de_dot('a.b.c', 'value'), {'a': {'b': {'c': 'value'}}})

    def test_deepmerge(self):
        """Test recursive merging of dictionaries."""
        source = {'a': {'b': {'c': 'value'}}}
        destination = {'a': {'b': {'d': 'other'}}}
        result = deepmerge(source, destination)
        self.assertEqual(result, {'a': {'b': {'c': 'value', 'd': 'other'}}})



"""Functions and classes used for tests"""

import os
import random
import shutil
import string
import tempfile
import click
from es_client.debug import debug
from es_client.defaults import LOGGING_SETTINGS
from es_client import config as cfgfn
from es_client.utils import option_wrapper, prune_nones

debug.level = 5

ONOFF = {"on": "", "off": "no-"}

DEFAULT_HOST = "http://127.0.0.1:9200"

DEFAULTCFG = "\n".join(
    ["---", "elasticsearch:", "  client:", f"    hosts: [{DEFAULT_HOST}]"]
)

EMPTYCFG = (
    "---\n"
    "elasticsearch:\n"
    "  client:\n"
    "    hosts: \n"
    "      - \n"
    "    cloud_id: \n"
)

TESTUSER = "joe_user"
TESTPASS = "password"

YAMLCONFIG = (
    "---\n"
    "elasticsearch:\n"
    "  client:\n"
    f"    hosts: [{DEFAULT_HOST}]\n"
    "  other_settings:\n"
    "    username: {0}\n"
    "    password: {1}\n"
)

click_opt_wrap = option_wrapper()


def random_directory():
    """Create a random dictionary"""
    dirname = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
    )
    directory = tempfile.mkdtemp(suffix=dirname)
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


class FileTestObj(object):
    """All file tests will use this object"""

    def __init__(self):
        self.args = {}
        dirname = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
        )
        filename = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
        )
        # This will create a psuedo-random temporary directory on the machine
        # which runs the unit tests, but NOT on the machine where elasticsearch
        # is running. This means tests may fail if run against remote instances
        # unless you explicitly set `self.args['location']` to a proper spot
        # on the target machine.
        self.written_value = """NOTHING"""
        self.args["tmpdir"] = tempfile.mkdtemp(suffix=dirname)
        if not os.path.exists(self.args["tmpdir"]):
            os.makedirs(self.args["tmpdir"])
        self.args["configdir"] = random_directory()
        self.args["configfile"] = os.path.join(self.args["configdir"], "es_client.yml")
        self.args["filename"] = os.path.join(self.args["tmpdir"], filename)
        self.args["no_file_here"] = os.path.join(self.args["tmpdir"], "not_created")
        with open(self.args["filename"], "w", encoding="utf-8") as f:
            f.write(self.written_value)

    def teardown(self):
        """Default teardown"""
        if os.path.exists(self.args["tmpdir"]):
            shutil.rmtree(self.args["tmpdir"])
        if os.path.exists(self.args["configdir"]):
            shutil.rmtree(self.args["configdir"])

    def write_config(self, fname, data):
        """Write config to named file"""
        with open(fname, "w", encoding="utf-8") as f:
            f.write(data)


# pylint: disable=unused-argument, redefined-builtin, too-many-arguments
@click.command()
@click_opt_wrap(*cfgfn.cli_opts("config"))
@click_opt_wrap(*cfgfn.cli_opts("hosts"))
@click_opt_wrap(*cfgfn.cli_opts("cloud_id"))
@click_opt_wrap(*cfgfn.cli_opts("api_token"))
@click_opt_wrap(*cfgfn.cli_opts("id"))
@click_opt_wrap(*cfgfn.cli_opts("api_key"))
@click_opt_wrap(*cfgfn.cli_opts("username"))
@click_opt_wrap(*cfgfn.cli_opts("password"))
@click_opt_wrap(*cfgfn.cli_opts("bearer_auth"))
@click_opt_wrap(*cfgfn.cli_opts("opaque_id"))
@click_opt_wrap(*cfgfn.cli_opts("request_timeout"))
@click_opt_wrap(*cfgfn.cli_opts("http_compress", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("verify_certs", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("ca_certs"))
@click_opt_wrap(*cfgfn.cli_opts("client_cert"))
@click_opt_wrap(*cfgfn.cli_opts("client_key"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_hostname"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_fingerprint"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_version"))
@click_opt_wrap(*cfgfn.cli_opts("master-only", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("skip_version_test", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("loglevel", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logfile", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logformat", settings=LOGGING_SETTINGS))
@click.pass_context
def simulator(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
):
    """Test command with all regular options"""
    ctx.obj = {}
    cfgfn.get_config(ctx)
    cfgfn.generate_configdict(ctx)
    click.echo(f'{ctx.obj["configdict"]}')


# pylint: disable=unused-argument
@click.command()
@click_opt_wrap(*cfgfn.cli_opts("config"))
@click_opt_wrap(*cfgfn.cli_opts("hosts"))
@click_opt_wrap(*cfgfn.cli_opts("cloud_id"))
@click_opt_wrap(*cfgfn.cli_opts("api_token"))
@click_opt_wrap(*cfgfn.cli_opts("id"))
@click_opt_wrap(*cfgfn.cli_opts("api_key"))
@click_opt_wrap(*cfgfn.cli_opts("username"))
@click_opt_wrap(*cfgfn.cli_opts("password"))
@click_opt_wrap(*cfgfn.cli_opts("bearer_auth"))
@click_opt_wrap(*cfgfn.cli_opts("opaque_id"))
@click_opt_wrap(*cfgfn.cli_opts("request_timeout"))
@click_opt_wrap(*cfgfn.cli_opts("http_compress", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("verify_certs", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("ca_certs"))
@click_opt_wrap(*cfgfn.cli_opts("client_cert"))
@click_opt_wrap(*cfgfn.cli_opts("client_key"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_hostname"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_fingerprint"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_version"))
@click_opt_wrap(*cfgfn.cli_opts("master-only", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("skip_version_test", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("loglevel", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logfile", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logformat", settings=LOGGING_SETTINGS))
@click.pass_context
def default_config_cmd(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
):
    """Test command with all regular options"""
    # Build config file
    file_obj = FileTestObj()
    file_obj.write_config(
        file_obj.args["configfile"], YAMLCONFIG.format(TESTUSER, TESTPASS)
    )
    # User config file
    ctx.obj = {"default_config": file_obj.args["configfile"]}
    cfgfn.get_config(ctx)
    # Teardown config file
    file_obj.teardown()
    # Finish the function
    cfgfn.generate_configdict(ctx)
    click.echo(f'{ctx.obj["configdict"]}')


@click.command()
@click_opt_wrap(*cfgfn.cli_opts("config"))
@click_opt_wrap(*cfgfn.cli_opts("hosts"))
@click_opt_wrap(*cfgfn.cli_opts("cloud_id"))
@click_opt_wrap(*cfgfn.cli_opts("api_token"))
@click_opt_wrap(*cfgfn.cli_opts("id"))
@click_opt_wrap(*cfgfn.cli_opts("api_key"))
@click_opt_wrap(*cfgfn.cli_opts("username"))
@click_opt_wrap(*cfgfn.cli_opts("password"))
@click_opt_wrap(*cfgfn.cli_opts("bearer_auth"))
@click_opt_wrap(*cfgfn.cli_opts("opaque_id"))
@click_opt_wrap(*cfgfn.cli_opts("request_timeout"))
@click_opt_wrap(*cfgfn.cli_opts("http_compress", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("verify_certs", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("ca_certs"))
@click_opt_wrap(*cfgfn.cli_opts("client_cert"))
@click_opt_wrap(*cfgfn.cli_opts("client_key"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_hostname"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_assert_fingerprint"))
@click_opt_wrap(*cfgfn.cli_opts("ssl_version"))
@click_opt_wrap(*cfgfn.cli_opts("master-only", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("skip_version_test", onoff=ONOFF))
@click_opt_wrap(*cfgfn.cli_opts("loglevel", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logfile", settings=LOGGING_SETTINGS))
@click_opt_wrap(*cfgfn.cli_opts("logformat", settings=LOGGING_SETTINGS))
@click.pass_context
def simulate_override_client_args(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
):
    """Test command with all regular options"""
    ctx.obj = {}
    cfgfn.get_config(ctx)
    cfgfn.get_arg_objects(ctx)
    # Manual override
    ctx.obj["client_args"].hosts = None
    cfgfn.override_client_args(ctx)
    ctx.obj["configdict"] = {
        "elasticsearch": {
            "client": prune_nones(ctx.obj["client_args"].toDict()),
            "other_settings": prune_nones(ctx.obj["other_args"].toDict()),
        }
    }
    click.echo(f'{ctx.obj["configdict"]}')
